import datetime
from pathlib import Path
from typing import List, Tuple

import pandas as pd
import pytest
from _pytest.fixtures import SubRequest

from webviz_subsurface._providers import EnsembleSummaryProvider
from webviz_subsurface._providers.ensemble_summary_provider._provider_impl_arrow_presampled import (
    ProviderImplArrowPresampled,
)
from webviz_subsurface.plugins._group_tree._utils._ensemble_group_tree_data import (
    add_nodetype,
)

ADD_NODETYPE_CASES = [
    # Group leaf nodes:
    # NODE1 has summary data>0 and will be classified as prod and inj
    # NODE2 has summary data==0 and will be classified as other
    # NODE3 has no summary data and will be classified as other
    # FIELD and TMPL are classified as all three types
    (
        pd.DataFrame(
            columns=["DATE", "CHILD", "KEYWORD", "PARENT"],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None],
                ["2000-01-01", "TMPL", "GRUPTREE", "FIELD"],
                ["2000-01-01", "NODE1", "GRUPTREE", "TMPL"],
                ["2000-01-01", "NODE2", "GRUPTREE", "TMPL"],
                ["2000-01-01", "NODE3", "GRUPTREE", "TMPL"],
            ],
        ),
        pd.DataFrame(
            columns=[
                "DATE",
                "REAL",
                "GGPR:NODE1",
                "GGIR:NODE1",
                "GGPR:NODE2",
                "GGIR:NODE2",
            ],
            data=[
                [datetime.date(2000, 1, 1), 0, 1, 1, 0, 0],
            ],
        ),
        pd.DataFrame(
            columns=[
                "DATE",
                "CHILD",
                "KEYWORD",
                "PARENT",
                "IS_PROD",
                "IS_INJ",
                "IS_OTHER",
            ],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None, True, True, True],
                ["2000-01-01", "TMPL", "GRUPTREE", "FIELD", True, True, True],
                ["2000-01-01", "NODE1", "GRUPTREE", "TMPL", True, True, False],
                ["2000-01-01", "NODE2", "GRUPTREE", "TMPL", False, False, True],
                ["2000-01-01", "NODE3", "GRUPTREE", "TMPL", False, False, True],
            ],
        ),
    ),
    # Well leaf nodes:
    # WELL1 has WSTAT==1 and will be classified as producer
    # WELL2 has WSTAT==2 and will be classified as injector
    # WELL3 has first WSTAT==1 and then WSTAT==2 and will be classified as both
    # WELL4 has WSTAT==0 and will be classified as other
    # TMPL_A is classified as prod and inj
    # TMPL_B is classified as other
    (
        pd.DataFrame(
            columns=["DATE", "CHILD", "KEYWORD", "PARENT"],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None],
                ["2000-01-01", "TMPL_A", "GRUPTREE", "FIELD"],
                ["2000-01-01", "TMPL_B", "GRUPTREE", "FIELD"],
                ["2000-01-01", "WELL1", "WELSPECS", "TMPL_A"],
                ["2000-01-01", "WELL2", "WELSPECS", "TMPL_A"],
                ["2000-01-01", "WELL3", "WELSPECS", "TMPL_A"],
                ["2000-01-01", "WELL4", "WELSPECS", "TMPL_B"],
            ],
        ),
        pd.DataFrame(
            columns=[
                "DATE",
                "REAL",
                "WSTAT:WELL1",
                "WSTAT:WELL2",
                "WSTAT:WELL3",
                "WSTAT:WELL4",
            ],
            data=[
                [datetime.date(2000, 1, 1), 0, 1, 2, 1, 0],
                [datetime.date(2000, 2, 1), 0, 1, 2, 2, 0],
            ],
        ),
        pd.DataFrame(
            columns=[
                "DATE",
                "CHILD",
                "KEYWORD",
                "PARENT",
                "IS_PROD",
                "IS_INJ",
                "IS_OTHER",
            ],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None, True, True, True],
                ["2000-01-01", "TMPL_A", "GRUPTREE", "FIELD", True, True, False],
                ["2000-01-01", "TMPL_B", "GRUPTREE", "FIELD", False, False, True],
                ["2000-01-01", "WELL1", "WELSPECS", "TMPL_A", True, False, False],
                ["2000-01-01", "WELL2", "WELSPECS", "TMPL_A", False, True, False],
                ["2000-01-01", "WELL3", "WELSPECS", "TMPL_A", True, True, False],
                ["2000-01-01", "WELL4", "WELSPECS", "TMPL_B", False, False, True],
            ],
        ),
    ),
]


@pytest.fixture(
    name="testdata",
    params=ADD_NODETYPE_CASES,
)
def fixture_provider(
    request: SubRequest, tmp_path: Path
) -> Tuple[pd.DataFrame, EnsembleSummaryProvider, pd.DataFrame]:

    input_py = request.param
    storage_dir = tmp_path
    gruptree_df = input_py[0]
    smry_df = input_py[1]
    expected_df = input_py[2]

    ProviderImplArrowPresampled.write_backing_store_from_ensemble_dataframe(
        storage_dir, "dummy_key", smry_df
    )
    new_provider = ProviderImplArrowPresampled.from_backing_store(
        storage_dir, "dummy_key"
    )

    if not new_provider:
        raise ValueError("Failed to create EnsembleSummaryProvider")

    return gruptree_df, new_provider, expected_df


def test_add_nodetype(
    testdata: Tuple[pd.DataFrame, EnsembleSummaryProvider, pd.DataFrame]
) -> None:
    """Test functionality for the add_nodetype function"""
    gruptree_df = testdata[0]
    provider = testdata[1]
    expected_df = testdata[2]

    columns_to_check = [
        "DATE",
        "CHILD",
        "KEYWORD",
        "PARENT",
        "IS_PROD",
        "IS_INJ",
        "IS_OTHER",
    ]

    wells: List[str] = gruptree_df[gruptree_df["KEYWORD"] == "WELSPECS"][
        "CHILD"
    ].unique()

    output = add_nodetype(gruptree_df, provider, wells, "FIELD")
    pd.testing.assert_frame_equal(
        output[columns_to_check], expected_df[columns_to_check]
    )



from typing import Any, Dict, List, Optional, Tuple

import webviz_core_components as wcc
from dash import Input, Output, State, callback, html
from dash.development.base_component import Component
from webviz_config import WebvizConfigTheme
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from ..._types import PressurePlotMode
from ..._utils import EnsembleWellAnalysisData
from ._utils import create_well_control_figure
from ._view_element import WellControlViewElement


class WellControlSettings(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLE = "ensemble"
        WELL = "well"
        SHARED_X_AXIS = "shared-x-axis"

    def __init__(self, data_models: Dict[str, EnsembleWellAnalysisData]) -> None:
        super().__init__("Plot Controls")
        self.ensembles = list(data_models.keys())
        self.wells: List[str] = []
        for _, ens_data_model in data_models.items():
            self.wells.extend(
                [well for well in ens_data_model.wells if well not in self.wells]
            )

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Ensemble",
                id=self.register_component_unique_id(WellControlSettings.Ids.ENSEMBLE),
                options=[{"label": col, "value": col} for col in self.ensembles],
                value=self.ensembles[0],
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Well",
                id=self.register_component_unique_id(WellControlSettings.Ids.WELL),
                options=[{"label": well, "value": well} for well in self.wells],
                value=self.wells[0],
                multi=False,
                clearable=False,
            ),
            wcc.Checklist(
                id=self.register_component_unique_id(
                    WellControlSettings.Ids.SHARED_X_AXIS
                ),
                options=[{"label": "Shared x-axis", "value": "shared_xaxes"}],
                value=["shared_xaxes"],
            ),
        ]


class WellControlPressurePlotOptions(SettingsGroupABC):
    class Ids(StrEnum):
        INCLUDE_BHP = "include-bhp"
        PRESSURE_PLOT_MODE = "pressure-plot-mode"
        REALIZATION_BOX = "realization-box"
        REALIZATION = "realization"
        DISPLAY_CTRL_MODE_BAR = "display-ctrl-mode-bar"

    def __init__(self, data_models: Dict[str, EnsembleWellAnalysisData]) -> None:

        super().__init__("Pressure Plot Options")
        self.data_models = data_models
        self.ensembles = list(data_models.keys())

    def layout(self) -> List[Component]:
        return [
            wcc.Checklist(
                id=self.register_component_unique_id(
                    WellControlPressurePlotOptions.Ids.INCLUDE_BHP
                ),
                options=[{"label": "Include BHP", "value": "include_bhp"}],
                value=["include_bhp"],
            ),
            wcc.RadioItems(
                label="Mean or realization",
                id=self.register_component_unique_id(
                    WellControlPressurePlotOptions.Ids.PRESSURE_PLOT_MODE
                ),
                options=[
                    {
                        "label": "Mean of producing real.",
                        "value": PressurePlotMode.MEAN,
                    },
                    {
                        "label": "Single realization",
                        "value": PressurePlotMode.SINGLE_REAL,
                    },
                ],
                value=PressurePlotMode.MEAN,
            ),
            html.Div(
                id=self.register_component_unique_id(
                    WellControlPressurePlotOptions.Ids.REALIZATION_BOX
                ),
                children=[
                    wcc.Dropdown(
                        id=self.register_component_unique_id(
                            WellControlPressurePlotOptions.Ids.REALIZATION
                        ),
                        options=[],
                        value=None,
                        multi=False,
                    ),
                    wcc.Checklist(
                        id=self.register_component_unique_id(
                            WellControlPressurePlotOptions.Ids.DISPLAY_CTRL_MODE_BAR
                        ),
                        options=[
                            {
                                "label": "Display ctrl mode bar",
                                "value": "ctrlmode_bar",
                            }
                        ],
                        value=["ctrlmode_bar"],
                    ),
                ],
            ),
        ]


class WellControlView(ViewABC):
    class Ids(StrEnum):
        SETTINGS = "settings"
        PRESSUREPLOT_OPTIONS = "pressure-plot-options"
        VIEW_ELEMENT = "view-element"

    def __init__(
        self,
        data_models: Dict[str, EnsembleWellAnalysisData],
        theme: WebvizConfigTheme,
    ) -> None:
        super().__init__("Well Control")

        self.data_models = data_models
        self.theme = theme

        self.add_settings_group(
            WellControlSettings(self.data_models), WellControlView.Ids.SETTINGS
        )
        self.add_settings_group(
            WellControlPressurePlotOptions(self.data_models),
            WellControlView.Ids.PRESSUREPLOT_OPTIONS,
        )

        main_column = self.add_column()
        main_column.add_view_element(WellControlViewElement(), self.Ids.VIEW_ELEMENT)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.settings_group(WellControlView.Ids.SETTINGS)
                .component_unique_id(WellControlSettings.Ids.WELL)
                .to_string(),
                "options",
            ),
            Output(
                self.settings_group(WellControlView.Ids.SETTINGS)
                .component_unique_id(WellControlSettings.Ids.WELL)
                .to_string(),
                "value",
            ),
            Output(
                self.settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(WellControlPressurePlotOptions.Ids.REALIZATION)
                .to_string(),
                "options",
            ),
            Output(
                self.settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(WellControlPressurePlotOptions.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellControlView.Ids.SETTINGS)
                .component_unique_id(WellControlSettings.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
            State(
                self.settings_group(WellControlView.Ids.SETTINGS)
                .component_unique_id(WellControlSettings.Ids.WELL)
                .to_string(),
                "value",
            ),
            State(
                self.settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(WellControlPressurePlotOptions.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
        )
        def _update_dropdowns(
            ensemble: str, state_well: str, state_real: int
        ) -> Tuple[
            List[Dict[str, str]], Optional[str], List[Dict[str, Any]], Optional[int]
        ]:
            """Updates the well and realization dropdowns with ensemble values"""
            wells = self.data_models[ensemble].wells
            reals = self.data_models[ensemble].realizations
            return (
                [{"label": well, "value": well} for well in wells],
                state_well if state_well in wells else wells[0],
                [{"label": real, "value": real} for real in reals],
                state_real if state_real in reals else reals[0],
            )

        @callback(
            Output(
                self.view_element(self.Ids.VIEW_ELEMENT)
                .component_unique_id(WellControlViewElement.Ids.CHART)
                .to_string(),
                "children",
            ),
            Input(
                self.settings_group(WellControlView.Ids.SETTINGS)
                .component_unique_id(WellControlSettings.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellControlView.Ids.SETTINGS)
                .component_unique_id(WellControlSettings.Ids.WELL)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(WellControlPressurePlotOptions.Ids.INCLUDE_BHP)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(
                    WellControlPressurePlotOptions.Ids.PRESSURE_PLOT_MODE
                )
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(WellControlPressurePlotOptions.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(
                    WellControlPressurePlotOptions.Ids.DISPLAY_CTRL_MODE_BAR
                )
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellControlView.Ids.SETTINGS)
                .component_unique_id(WellControlSettings.Ids.SHARED_X_AXIS)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _update_graph(
            ensemble: str,
            well: str,
            include_bhp: List[str],
            pressure_plot_mode: PressurePlotMode,
            real: int,
            display_ctrlmode_bar: List[str],
            shared_xaxes: List[str],
        ) -> Component:
            """Updates the well control figure"""
            fig = create_well_control_figure(
                self.data_models[ensemble].get_node_info(
                    well, pressure_plot_mode, real
                ),
                self.data_models[ensemble].summary_data,
                pressure_plot_mode,
                real,
                "ctrlmode_bar" in display_ctrlmode_bar,
                "shared_xaxes" in shared_xaxes,
                "include_bhp" in include_bhp,
                self.theme,
            )

            return wcc.Graph(style={"height": "87vh"}, figure=fig)

        @callback(
            Output(
                self.settings_group(self.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(WellControlPressurePlotOptions.Ids.REALIZATION_BOX)
                .to_string(),
                component_property="style",
            ),
            Input(
                self.settings_group(self.Ids.PRESSUREPLOT_OPTIONS)
                .component_unique_id(
                    WellControlPressurePlotOptions.Ids.PRESSURE_PLOT_MODE
                )
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _show_hide_single_real_options(
            pressure_plot_mode: PressurePlotMode,
        ) -> Dict[str, str]:
            """Hides or unhides the realization dropdown according to whether mean
            or single realization is selected.
            """
            if pressure_plot_mode == PressurePlotMode.MEAN:
                return {"display": "none"}
            return {"display": "block"}



from webviz_config.utils import StrEnum


class LineType(StrEnum):
    REALIZATION = "realization"
    FANCHART = "fanchart"


class DepthType(StrEnum):
    TVD = "TVD"
    MD = "MD"


class ColorAndSizeByType(StrEnum):
    MISFIT = "ABSDIFF"
    STDDEV = "STDDEV"
    YEAR = "YEAR"


class CorrType(StrEnum):
    SIM_VS_PARAM = "sim_vs_param"
    PARAM_VS_SIM = "param_vs_sim"



import logging
from pathlib import Path

from .ensemble_surface_provider import (
    EnsembleSurfaceProvider,
    ObservedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceStatistic,
)
from .ensemble_surface_provider_factory import EnsembleSurfaceProviderFactory


def main() -> None:
    print()
    print("## Running EnsembleSurfaceProvider experiments")
    print("## =================================================")

    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s %(levelname)-3s [%(name)s]: %(message)s",
    )
    logging.getLogger("webviz_subsurface").setLevel(level=logging.DEBUG)

    root_storage_dir = Path("/home/sigurdp/buf/webviz_storage_dir")

    # fmt:off
    # ensemble_path = "../webviz-subsurface-testdata/01_drogon_ahm/realization-*/iter-0"
    ensemble_path = "../hk-webviz-subsurface-testdata/01_drogon_ahm/realization-*/iter-0"
    # fmt:on

    # WEBVIZ_FACTORY_REGISTRY.initialize(
    #     WebvizInstanceInfo(WebvizRunMode.NON_PORTABLE, root_storage_dir), None
    # )
    # factory = EnsembleSurfaceProviderFactory.instance()

    factory = EnsembleSurfaceProviderFactory(
        root_storage_dir, allow_storage_writes=True, avoid_copying_surfaces=False
    )

    provider: EnsembleSurfaceProvider = factory.create_from_ensemble_surface_files(
        ensemble_path
    )

    all_attributes = provider.attributes()
    print()
    print("all_attributes:")
    print("------------------------")
    print(*all_attributes, sep="\n")

    print()
    print("attributes for names:")
    print("------------------------")
    for attr in all_attributes:
        print(f"attr={attr}:")
        print(f"  surf_names={provider.surface_names_for_attribute(attr)}")
        print(f"  surf_dates={provider.surface_dates_for_attribute(attr)}")

    print()
    all_realizations = provider.realizations()
    print(f"all_realizations={all_realizations}")

    surf = provider.get_surface(
        SimulatedSurfaceAddress(
            attribute="oilthickness",
            name="therys",
            datestr="20200701_20180101",
            realization=1,
        )
    )
    print(surf)

    surf = provider.get_surface(
        ObservedSurfaceAddress(
            attribute="amplitude_mean",
            name="basevolantis",
            datestr="20180701_20180101",
        )
    )
    print(surf)

    # surf = provider.get_surface(
    #     StatisticalSurfaceAddress(
    #         attribute="amplitude_mean",
    #         name="basevolantis",
    #         datestr="20180701_20180101",
    #         statistic=SurfaceStatistic.P10,
    #         realizations=[0, 1],
    #     )
    # )
    # print(surf)

    # surf = provider.get_surface(
    #     StatisticalSurfaceAddress(
    #         attribute="amplitude_mean",
    #         name="basevolantis",
    #         datestr="20180701_20180101",
    #         statistic=SurfaceStatistic.P10,
    #         realizations=all_realizations,
    #     )
    # )
    # print(surf)

    surf = provider.get_surface(
        StatisticalSurfaceAddress(
            attribute="ds_extract_postprocess-refined8",
            name="topvolantis",
            datestr=None,
            statistic=SurfaceStatistic.P10,
            realizations=all_realizations,
        )
    )
    print(surf)


# Running:
#   python -m webviz_subsurface._providers.ensemble_surface_provider.dev_experiments
# -------------------------------------------------------------------------
if __name__ == "__main__":
    main()



from typing import List

import webviz_core_components as wcc
from dash import html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from webviz_subsurface._providers import EnsembleTableProvider


class Filters(SettingsGroupABC):
    class IDs(StrEnum):
        SINGLE_FILTER = "single-filter"
        MULTI_FILTER = "multi-filter"

    def __init__(
        self,
        table_provider: EnsembleTableProvider,
        single_filters: List[str],
        multi_filters: List[str],
    ) -> None:
        super().__init__("Filters")
        self._table_provider = table_provider
        self._single_filters = single_filters
        self._multi_filters = multi_filters
        self.single_filter_id = self.register_component_unique_id(
            Filters.IDs.SINGLE_FILTER
        )
        self.multi_filter_id = self.register_component_unique_id(
            Filters.IDs.MULTI_FILTER
        )

    def layout(self) -> List[Component]:
        single_filter_elements = []

        for selector in self._single_filters:
            values = self._table_provider.get_column_data([selector])[selector].unique()

            single_filter_elements.append(
                wcc.Dropdown(
                    label=selector,
                    id={
                        "id": self.single_filter_id,
                        "name": selector,
                        "type": "single_filter",
                    },
                    options=[{"label": val, "value": val} for val in values],
                    value=values[0],
                    clearable=False,
                )
            )

        multi_filter_elements = []
        for selector in self._multi_filters:
            values = self._table_provider.get_column_data([selector])[selector].unique()
            multi_filter_elements.append(
                wcc.SelectWithLabel(
                    label=selector,
                    id={
                        "id": self.multi_filter_id,
                        "name": selector,
                        "type": "multi_filter",
                    },
                    options=[{"label": val, "value": val} for val in values],
                    value=values,
                    multi=True,
                )
            )
        return [
            html.Div(
                id=self.component_unique_id(self.single_filter_id).to_string(),
                children=single_filter_elements,
            ),
            html.Div(
                id=self.component_unique_id(self.multi_filter_id).to_string(),
                children=multi_filter_elements,
            ),
        ]



import datetime
from enum import Enum
from typing import List, Optional

import numpy as np
import pandas as pd

from webviz_subsurface._utils.colors import find_intermediate_color, rgba_to_str
from webviz_subsurface._utils.simulation_timeseries import (
    get_simulation_line_shape,
    set_simulation_line_shape_fallback,
)


class Colors(str, Enum):
    RED = rgba_to_str((255, 18, 67, 1))
    MID = rgba_to_str((220, 220, 220, 1))
    GREEN = rgba_to_str((62, 208, 62, 1))


class TimeSeriesFigure:
    STAT_OPTIONS = ["Mean", "P10", "P90"]
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        dframe: pd.DataFrame,
        visualization: str,
        vector: str,
        ensemble: str,
        color_col: str,
        line_shape_fallback: str,
        historical_vector_df: Optional[pd.DataFrame] = None,
        dateline: Optional[datetime.datetime] = None,
    ):
        self.dframe = dframe
        self.vector = vector
        self.ensemble = ensemble
        self.color_col = color_col
        self.visualization = visualization
        self.historical_vector_df = historical_vector_df
        self.date = dateline
        self.line_shape = self.get_line_shape(line_shape_fallback)

        self.create_traces()

    @property
    def figure(self) -> dict:
        return {
            "data": self.traces,
            "layout": {
                "margin": {"r": 40, "l": 20, "t": 60, "b": 20},
                "yaxis": {"automargin": True, "gridcolor": "#ECECEC"},
                "xaxis": {"range": self.daterange, "gridcolor": "#ECECEC"},
                "hovermode": "closest",
                "paper_bgcolor": "white",
                "plot_bgcolor": "white",
                "showlegend": False,
                "uirevision": self.vector,
                "title": {"text": f"{self.vector} colored by {self.color_col}"},
                "shapes": self.shapes,
                "annotations": self.annotations,
            },
        }

    def create_traces(self) -> None:
        self.traces: List[dict] = []

        if self.visualization != "statistics":
            self._add_realization_traces()

        if self.visualization != "realizations":
            self._add_statistic_traces()

        self._add_history_trace()

    def _add_history_trace(self) -> None:
        """Renders the history line"""
        if self.historical_vector_df is None:
            return

        self.traces.append(
            {
                "line": {"shape": self.line_shape, "color": "black"},
                "x": self.historical_vector_df["DATE"],
                "y": self.historical_vector_df[self.vector],
                "mode": "lines",
                "hovertext": "History",
                "hoverinfo": "y+x+text",
                "name": "History",
            }
        )

    def _add_statistic_traces(self) -> None:
        """Renders the statistic lines"""
        stat_df = self.create_vectors_statistics_df()
        stat_df = pd.DataFrame(stat_df["DATE"]).join(stat_df[self.vector])
        self.traces.extend(
            [
                {
                    "line": {
                        "width": 2,
                        "shape": self.line_shape,
                        "dash": False if stat == "Mean" else "dashdot",
                        "color": Colors.RED,
                    },
                    "mode": "lines",
                    "x": stat_df["DATE"],
                    "y": stat_df[stat],
                    "name": stat,
                    "legendgroup": self.ensemble,
                    "showlegend": False,
                }
                for stat in self.STAT_OPTIONS
            ]
        )

    def _add_realization_traces(self) -> None:
        """Renders line trace for each realization"""
        mean = self.dframe["VALUE_NORM"].mean()
        self.traces.extend(
            [
                {
                    "line": {
                        "shape": self.line_shape,
                        "color": self.set_real_color(
                            real_df["VALUE_NORM"].iloc[0], mean
                        )
                        if self.visualization == "realizations"
                        else "gainsboro",
                    },
                    "mode": "lines",
                    "x": real_df["DATE"],
                    "y": real_df[self.vector],
                    "name": self.ensemble,
                    "legendgroup": self.ensemble,
                    "hovertext": self.create_hovertext(real_df["VALUE"].iloc[0], real),
                    "showlegend": real_idx == 0,
                }
                for real_idx, (real, real_df) in enumerate(self.dframe.groupby("REAL"))
            ]
        )

    @property
    def daterange(self) -> list:
        active_dates = self.dframe["DATE"][self.dframe[self.vector] != 0]
        if len(active_dates) == 0:
            return [self.dframe["DATE"].min(), self.dframe["DATE"].max()]
        if self.date is None:
            return [active_dates.min(), active_dates.max()]
        # Ensure xaxis covers selected date
        return [min(active_dates.min(), self.date), max(active_dates.max(), self.date)]

    @property
    def annotations(self) -> List[dict]:
        return (
            [
                {
                    "bgcolor": "white",
                    "showarrow": False,
                    "text": self.date.strftime("%Y-%m-%d"),
                    "x": self.date,
                    "y": 1,
                    "yref": "y domain",
                }
            ]
            if self.date is not None
            else []
        )

    @property
    def shapes(self) -> List[dict]:
        return (
            [
                {
                    "line": {"color": "#243746", "dash": "dot", "width": 4},
                    "type": "line",
                    "x0": self.date,
                    "x1": self.date,
                    "xref": "x",
                    "y0": 0,
                    "y1": 1,
                    "yref": "y domain",
                }
            ]
            if self.date is not None
            else []
        )

    def create_hovertext(self, color_value: float, real: int) -> str:
        return f"Real: {real}, {self.color_col}: {color_value}"

    def get_line_shape(self, line_shape_fallback: str) -> str:
        return get_simulation_line_shape(
            line_shape_fallback=set_simulation_line_shape_fallback(line_shape_fallback),
            vector=self.vector,
            smry_meta=None,
        )

    def create_vectors_statistics_df(self) -> pd.DataFrame:
        return (
            self.dframe[["DATE", self.vector]]
            .groupby(["DATE"])
            .agg(
                [
                    ("Mean", np.nanmean),
                    ("P10", lambda x: np.nanpercentile(x, q=90)),
                    ("P90", lambda x: np.nanpercentile(x, q=10)),
                ]
            )
            .reset_index()
        )

    @staticmethod
    def set_real_color(norm_value: float, mean_param_value: float) -> str:
        """
        Return color for trace based on normalized color_col value.
        Midpoint for the colorscale is set on the average value
        """

        if norm_value <= mean_param_value:
            intermed = norm_value / mean_param_value
            return find_intermediate_color(Colors.RED, Colors.MID, intermed)

        if norm_value > mean_param_value:
            intermed = (norm_value - mean_param_value) / (1 - mean_param_value)
            return find_intermediate_color(Colors.MID, Colors.GREEN, intermed)

        return Colors.MID



from typing import Optional

import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import html


def distributions_main_layout(uuid: str) -> html.Div:
    return html.Div(
        children=[
            html.Div(
                id={"id": uuid, "page": "custom"},
                style={"display": "block"},
            ),
            html.Div(
                id={"id": uuid, "page": "per_zr"},
                style={"display": "none"},
            ),
            html.Div(
                id={"id": uuid, "page": "conv"},
                style={"display": "none"},
            ),
        ]
    )


def table_main_layout(uuid: str) -> wcc.Frame:
    return html.Div(id={"id": uuid, "page": "table"})


def convergence_plot_layout(figure: go.Figure) -> wcc.Graph:
    return wcc.Graph(
        config={"displayModeBar": False},
        style={"height": "86vh"},
        figure=figure,
    )


def custom_plotting_layout(figure: go.Figure, tables: Optional[list]) -> html.Div:
    tables = tables if tables is not None else []
    return html.Div(
        children=[
            wcc.Graph(
                config={"displayModeBar": False},
                style={"height": "45vh" if tables else "86vh"},
                figure=figure,
            )
        ]
        + [html.Div(table, style={"margin-top": "20px"}) for table in tables]
    )


def plots_per_zone_region_layout(figures: list) -> list:
    height = "42vh" if len(figures) < 3 else "28vh"
    return html.Div(
        children=[
            html.Div(
                "Pie chart available if no 'Color by' is selected",
                style={"text-align": "right"},
            )
        ]
        + [
            wcc.FlexBox(
                style={"height": height},
                children=[
                    html.Div(
                        style={"flex": 3},
                        children=wcc.Graph(
                            config={"displayModeBar": False},
                            style={"height": height},
                            figure=barfig,
                        ),
                    ),
                    html.Div(
                        style={
                            "flex": 1,
                            "display": "block" if piefig else "none",
                        },
                        children=wcc.Graph(
                            config={"displayModeBar": False},
                            style={"height": height},
                            figure=piefig,
                        ),
                    ),
                ],
            )
            for piefig, barfig in figures
        ]
    )



import glob
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from fmu.ensemble import ScratchEnsemble


@dataclass(frozen=True)
class GridParameterFileInfo:
    path: str
    real: int
    name: str
    attribute: str
    datestr: Optional[str]


@dataclass(frozen=True)
class GridParameterIdent:
    name: str
    attribute: str
    datestr: Optional[str]


@dataclass(frozen=True)
class GridFileInfo:
    path: str
    real: int
    name: str


@dataclass(frozen=True)
class GridIdent:
    name: str


def _discover_ensemble_realizations_fmu(ens_path: str) -> Dict[int, str]:
    """Returns dict indexed by realization number and with runpath as value"""
    scratch_ensemble = ScratchEnsemble("dummyEnsembleName", paths=ens_path).filter("OK")
    real_dict = {i: r.runpath() for i, r in scratch_ensemble.realizations.items()}
    return real_dict


def _discover_ensemble_realizations(ens_path: str) -> Dict[int, str]:
    # Much faster than FMU impl above, but is it risky?
    # Do we need to check for OK-file?
    real_dict: Dict[int, str] = {}

    realidxregexp = re.compile(r"realization-(\d+)")
    globbed_real_dirs = sorted(glob.glob(str(ens_path)))
    for real_dir in globbed_real_dirs:
        realnum: Optional[int] = None
        for path_comp in reversed(real_dir.split(os.path.sep)):
            realmatch = re.match(realidxregexp, path_comp)
            if realmatch:
                realnum = int(realmatch.group(1))
                break

        if realnum is not None:
            real_dict[realnum] = real_dir

    return real_dict


def ident_from_filename(
    filename: str,
) -> Union[GridIdent, GridParameterIdent]:
    """Split the stem part of the roff filename into grid name, attribute and
    optionally date part"""
    delimiter: str = "--"
    parts = Path(filename).stem.split(delimiter)
    if len(parts) == 1:
        return GridIdent(name=parts[0])

    return GridParameterIdent(
        name=parts[0], attribute=parts[1], datestr=parts[2] if len(parts) >= 3 else None
    )


def discover_per_realization_roff_files(
    ens_path: str, grid_name: str, attribute_filter: List[str] = None
) -> Tuple[List[GridFileInfo], List[GridParameterFileInfo]]:
    rel_folder: str = "share/results/grids"
    suffix: str = "*.roff"

    grid_parameters_info: List[GridParameterFileInfo] = []
    grid_info: List[GridFileInfo] = []
    real_dict = _discover_ensemble_realizations_fmu(ens_path)
    for realnum, runpath in sorted(real_dict.items()):
        globbed_filenames = glob.glob(str(Path(runpath) / rel_folder / suffix))
        for filename in sorted(globbed_filenames):
            ident = ident_from_filename(filename)
            if ident.name != grid_name:
                continue
            if isinstance(ident, GridParameterIdent):
                if (
                    attribute_filter is not None
                    and ident.attribute not in attribute_filter
                ):
                    continue
                grid_parameters_info.append(
                    GridParameterFileInfo(
                        path=filename,
                        real=realnum,
                        name=ident.name,
                        attribute=ident.attribute,
                        datestr=ident.datestr,
                    )
                )
            else:
                grid_info.append(
                    GridFileInfo(path=filename, real=realnum, name=ident.name)
                )

    return grid_info, grid_parameters_info



import json
from pathlib import Path
from typing import Callable, Dict, List, Tuple, Union

import webviz_core_components as wcc
from dash import Dash, html
from webviz_config import WebvizPluginABC
from webviz_subsurface_components import WellLogViewer as WellLogViewerComponent

from webviz_subsurface._models.well_set_model import WellSetModel
from webviz_subsurface._utils.webvizstore_functions import find_files, get_path

from ._validate_log_templates import load_and_validate_log_templates
from .controllers import well_controller
from .utils.default_color_tables import default_color_tables
from .utils.xtgeo_well_log_to_json import xtgeo_well_logs_to_json_format


class WellLogViewer(WebvizPluginABC):
    """Uses [videx-welllog](https://github.com/equinor/videx-wellog) to visualize well logs
    from files stored in RMS well format.

?> Currently tracks for visualizing discrete logs are not included. This will
be added in later releases.

---

* **`wellfolder`:** Path to a folder with well files stored in RMS well format.
* **`wellsuffix`:** File suffix of well files
* **`logtemplates`:** List of yaml based log template configurations. \
    See the data section for description of the format.
* **`mdlog`:** Name of the md log. If not specified, MD will be calculated.
* **`well_tvdmin`:** Truncate well data values above this depth.
* **`well_tvdmax`:** Truncate well data values below this depth.
* **`well_downsample_interval`:** Sampling interval used for coarsening a well trajectory
* **`colortables`:** Color tables on json format. See https://git.io/JDLyb \
    for an example file.
* **`initial_settings`:** Configuration for initializing the plugin with various \
    properties set. All properties are optional.
    See the data section for available properties.

---

?> The format and documentation of the log template configuration will be improved \
in later releases. A small configuration sample is provided below.

```yaml
name: All logs # Name of the log template
scale:
  primary: MD # Which reference track to visualize as default (MD/TVD)
  allowSecondary: False # Set to True to show both MD and TVD reference tracks.
tracks: # The list of log tracks
  - title: Porosity # Optional title of log track
    plots: # List of which logs to include in the track
      - name: PHIT # Upper case name of log
        type: area # Type of visualiation (area, line, linestep, dot)
        color: green # Color of log
      - name: PHIT_ORIG
        type: line
  - plots:
      - name: ZONE
        type: area
  - plots:
      - name: FACIES
        type: area
  - plots:
      - name: VSH
        type: area
  - plots:
      - name: SW
        type: dot
styles: # List of styles that can be added to tracks
```


Format of the `initial_settings` argument:
```yaml
        initial_settings:
            well: str # Name of well
            logtemplate: str # Name of log template
```
"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        wellfolder: Path,
        logtemplates: List[Path],
        colortables: Path = None,
        wellsuffix: str = ".w",
        mdlog: str = None,
        well_tvdmin: Union[int, float] = None,
        well_tvdmax: Union[int, float] = None,
        well_downsample_interval: int = None,
        initial_settings: Dict = None,
    ):

        super().__init__()
        self._wellfolder = wellfolder
        self._wellsuffix = wellsuffix
        self._logtemplatefiles = logtemplates
        self._wellfiles: List = json.load(
            find_files(folder=self._wellfolder, suffix=self._wellsuffix)
        )
        self._log_templates = load_and_validate_log_templates(
            [get_path(fn) for fn in self._logtemplatefiles]
        )
        self._well_set_model = WellSetModel(
            self._wellfiles,
            mdlog=mdlog,
            tvdmin=well_tvdmin,
            tvdmax=well_tvdmax,
            downsample_interval=well_downsample_interval,
        )
        self.colortable_file = colortables
        if self.colortable_file:
            self.colortables = json.loads(get_path(self.colortable_file).read_text())
        else:
            self.colortables = default_color_tables()

        initial_settings = initial_settings if initial_settings else {}
        self.initial_well_name = initial_settings.get(
            "well_name", self._well_set_model.well_names[0]
        )
        self.initial_log_template = initial_settings.get(
            "logtemplate", list(self._log_templates.keys())[0]
        )
        self.set_callbacks(app)

    @property
    def layout(self) -> html.Div:
        return wcc.FlexBox(
            [
                wcc.Frame(
                    style={"height": "90vh", "flex": 1},
                    children=[
                        wcc.Dropdown(
                            label="Well",
                            id=self.uuid("well"),
                            options=[
                                {"label": name, "value": name}
                                for name in self._well_set_model.well_names
                            ],
                            value=self.initial_well_name,
                            clearable=False,
                        ),
                        wcc.Dropdown(
                            label="Log template",
                            id=self.uuid("template"),
                            options=[
                                {"label": name, "value": name}
                                for name in list(self._log_templates.keys())
                            ],
                            value=self.initial_log_template,
                            clearable=False,
                        ),
                    ],
                ),
                wcc.Frame(
                    style={"flex": 6, "height": "90vh"},
                    children=[
                        WellLogViewerComponent(
                            id=self.uuid("well-log-viewer"),
                            template=self._log_templates.get(self.initial_log_template),
                            welllog=xtgeo_well_logs_to_json_format(
                                well=self._well_set_model.get_well(
                                    self.initial_well_name
                                )
                            ),
                            colorTables=self.colortables,
                        )
                    ],
                ),
            ]
        )

    def set_callbacks(self, app: Dash) -> None:
        well_controller(
            app=app,
            well_set_model=self._well_set_model,
            log_templates=self._log_templates,
            get_uuid=self.uuid,
        )

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        store_functions = [
            (find_files, [{"folder": self._wellfolder, "suffix": self._wellsuffix}])
        ]

        store_functions.extend([(get_path, [{"path": fn}]) for fn in self._wellfiles])
        store_functions.extend(
            [(get_path, [{"path": fn}]) for fn in self._logtemplatefiles]
        )
        if self.colortable_file is not None:
            store_functions.append((get_path, [{"path": self.colortable_file}]))
        return store_functions



from typing import Callable, Dict, Optional, Union

import webviz_core_components as wcc

from ..models import (
    PropertyStatisticsModel,
    ProviderTimeSeriesDataModel,
    SimulationTimeSeriesModel,
)
from .property_delta_view import property_delta_view
from .property_qc_view import property_qc_view
from .property_response_view import property_response_view


def main_view(
    get_uuid: Callable,
    property_model: PropertyStatisticsModel,
    vector_model: Optional[
        Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel]
    ],
    surface_folders: Optional[Dict] = None,
) -> wcc.Tabs:
    tabs = [
        wcc.Tab(
            label="Property QC",
            children=property_qc_view(get_uuid=get_uuid, property_model=property_model),
        )
    ]
    if len(property_model.ensembles) > 1:
        tabs.append(
            wcc.Tab(
                label="AHM impact on property",
                children=property_delta_view(
                    get_uuid=get_uuid,
                    property_model=property_model,
                    surface_folders=surface_folders,
                ),
            )
        )
    if vector_model is not None:
        tabs.append(
            wcc.Tab(
                label="Property impact on simulation profiles",
                children=property_response_view(
                    get_uuid=get_uuid,
                    property_model=property_model,
                    vector_model=vector_model,
                    surface_folders=surface_folders,
                ),
            ),
        )

    return wcc.Tabs(style={"width": "100%"}, children=tabs)



from typing import Callable, List, Union

import webviz_core_components as wcc
from dash import ALL, Dash, Input, Output, dash_table

from ..models import PropertyStatisticsModel


def property_qc_controller(
    get_uuid: Callable, property_model: PropertyStatisticsModel, app: Dash
) -> None:
    @app.callback(
        Output(get_uuid("property-qc-wrapper"), "children"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "qc"}, "value"),
        Input({"id": get_uuid("property-selector"), "tab": "qc"}, "value"),
        Input(
            {"id": get_uuid("filter-selector"), "tab": "qc", "selector": ALL},
            "value",
        ),
        Input(get_uuid("property-qc-plot-type"), "value"),
        Input(get_uuid("property-qc-axis-match"), "value"),
    )
    def _update_bars(
        ensembles: Union[str, List[str]],
        prop: str,
        selectors: list,
        plot_type: str,
        match_axis: List[str],
    ) -> Union[dash_table.DataTable, wcc.Graph]:
        ensembles = ensembles if isinstance(ensembles, list) else [ensembles]
        if plot_type == "table":
            columns, dframe = property_model.make_statistics_table(
                prop=prop, ensembles=ensembles, selector_values=selectors
            )
            return dash_table.DataTable(
                style_table={
                    "height": "75vh",
                    "overflow": "auto",
                    "fontSize": 15,
                },
                style_cell={"textAlign": "center"},
                style_cell_conditional=[
                    {"if": {"column_id": "label|"}, "textAlign": "left"}
                ],
                columns=columns,
                data=dframe,
                sort_action="native",
                filter_action="native",
                merge_duplicate_headers=True,
            )
        return wcc.Graph(
            id=get_uuid("property-qc-graph"),
            config={"displayModeBar": False},
            style={"height": "75vh"},
            figure=property_model.make_grouped_plot(
                prop=prop,
                ensembles=ensembles,
                selector_values=selectors,
                plot_type=plot_type,
                match_axis=bool(match_axis),
            ),
        )



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class FilterLayout(SettingsGroupABC):
    class Ids(StrEnum):
        FILTER_WELLS = "filter-wells"
        FILTER_ZONES = "filter-zones"
        FILTER_DATES = "filter-dates"

    def __init__(self, wells: List[str], zones: List[str], dates: List[str]) -> None:
        super().__init__("Filters")
        self._wells = wells
        self._zones = zones
        self._dates = dates

    def layout(self) -> List[Component]:
        return [
            wcc.SelectWithLabel(
                label="Wells",
                size=min(10, len(self._wells)),
                id=self.register_component_unique_id(self.Ids.FILTER_WELLS),
                options=[{"label": name, "value": name} for name in self._wells],
                value=self._wells,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Zones",
                size=min(10, len(self._zones)),
                id=self.register_component_unique_id(self.Ids.FILTER_ZONES),
                options=[{"label": name, "value": name} for name in self._zones],
                value=self._zones,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Dates",
                size=min(10, len(self._dates)),
                id=self.register_component_unique_id(self.Ids.FILTER_DATES),
                options=[{"label": name, "value": name} for name in self._dates],
                value=self._dates,
                multi=True,
            ),
        ]



from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple, Union
from uuid import uuid4

import webviz_core_components as wcc
from dash import Dash, Input, Output, State, dcc, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC
from webviz_config.deprecation_decorators import deprecated_plugin
from webviz_config.webviz_store import webvizstore
from webviz_subsurface_components import LeafletMap

from webviz_subsurface._models import SurfaceLeafletModel

from .._datainput.seismic import load_cube_data
from .._datainput.surface import load_surface
from .._datainput.well import load_well, make_well_layer
from .._datainput.xsection import XSectionFigure


@deprecated_plugin(
    "Relevant functionality is implemented in the StructuralUncertainty plugin."
)
class WellCrossSection(WebvizPluginABC):
    """Displays a cross section along a well with intersected surfaces,
and optionally seismic cubes.

!> See also WellCrossSectionFMU for additional functionality with FMU ensembles.
---

* **`segyfiles`:** List of file paths to segyfiles (absolute or relative to config file).
* **`surfacefiles`:** List of file paths to Irap binary surfaces \
(absolute or relative to config file).
* **`surfacenames`:** Corresponding list of displayed surface names.
* **`wellfiles`:** List of file paths to RMS wells (absolute or relative to config file).
* **`zunit`:** z-unit for display.
* **`zonelog`:** Name of zonelog (for the RMS wells in `wellfiles`).
* **`zmin`:** Visualized minimum z-value in cross section.
* **`zmax`:** Visualized maximum z-value in cross section.
* **`zonemin`:** First zonenumber to draw in log.
* **`sampling`:** Sampling interval of well fence.
* **`nextend`:** Number of samples to extend well fence on each side of well, \
e.g. with distance of sampling=20 and nextend=2: extension=2*20 (nextend*sampling). \

---

**Example files**

* [Segyfiles](https://github.com/equinor/webviz-subsurface-testdata/tree/master/\
observed_data/seismic).

* [One file for surfacefiles](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/\
maps/topupperreek--ds_extracted_horizons.gri).

* [Wellfiles](https://github.com/equinor/webviz-subsurface-testdata/tree/master/\
observed_data/wells).

The segyfiles are on a `SEG-Y` format and can be investigated outside `webviz` using \
e.g. [xtgeo](https://xtgeo.readthedocs.io/en/latest/).

The surfacefiles are on a `ROFF binary` format and can be investigated outside `webviz` using \
e.g. [xtgeo](https://xtgeo.readthedocs.io/en/latest/).

"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        surfacefiles: List[Path],
        wellfiles: List[Path],
        segyfiles: List[Path] = None,
        surfacenames: list = None,
        zonelog: str = None,
        zunit: str = "depth (m)",
        zmin: float = None,
        zmax: float = None,
        zonemin: int = 1,
        nextend: int = 2,
        sampling: int = 40,
    ):

        super().__init__()

        self.zunit = zunit
        self.sampling = sampling
        self.nextend = nextend
        self.zmin = zmin
        self.zmax = zmax
        self.zonemin = zonemin
        self.surfacefiles = [str(surface) for surface in surfacefiles]
        self.wellfiles = [str(well) for well in wellfiles]
        self.segyfiles = [str(segy) for segy in segyfiles] if segyfiles else []
        self.surfacenames: List[str] = (
            surfacenames if surfacenames else [surface.stem for surface in surfacefiles]
        )
        self.zonelog = zonelog
        self.uid = uuid4()
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def well_layout(self) -> html.Div:
        return html.Div(
            children=html.Label(
                children=[
                    html.Span("Well:", style={"font-weight": "bold"}),
                    dcc.Dropdown(
                        id=self.ids("wells"),
                        options=[
                            {"label": Path(well).stem, "value": well}
                            for well in self.wellfiles
                        ],
                        value=self.wellfiles[0],
                        clearable=False,
                        persistence=True,
                        persistence_type="session",
                    ),
                ]
            ),
        )

    @property
    def seismic_layout(self) -> html.Div:
        return html.Div(
            style={} if self.segyfiles else {"display": "none"},
            children=html.Label(
                children=[
                    html.Span("Seismic:", style={"font-weight": "bold"}),
                    dcc.Dropdown(
                        id=self.ids("cube"),
                        options=[
                            {"label": Path(segy).stem, "value": segy}
                            for segy in self.segyfiles
                        ]
                        if self.segyfiles
                        else None,
                        value=self.segyfiles[0] if self.segyfiles else None,
                        clearable=False,
                        persistence=True,
                        persistence_type="session",
                    ),
                ]
            ),
        )

    @property
    def viz_options_layout(self) -> dcc.Checklist:
        options = [{"label": "Show surface fill", "value": "show_surface_fill"}]
        value = ["show_surface_fill"]
        if self.segyfiles:
            options.append({"label": "Show seismic", "value": "show_seismic"})
        if self.zonelog:
            options.append({"label": "Show zonelog", "value": "show_zonelog"})
            value.append("show_zonelog")

        return dcc.Checklist(
            id=self.ids("options"),
            options=options,
            value=value,
            persistence=True,
            persistence_type="session",
        )

    @property
    def well_options(self) -> html.Div:
        return html.Div(
            style={"marginLeft": "20px", "marginRight": "0px", "marginBotton": "0px"},
            children=[
                html.Div(
                    children=html.Label(
                        children=[
                            html.Span("Sampling:", style={"font-weight": "bold"}),
                            dcc.Input(
                                id=self.ids("sampling"),
                                debounce=True,
                                type="number",
                                required=True,
                                value=self.sampling,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ]
                    )
                ),
                html.Div(
                    children=html.Label(
                        children=[
                            html.Span("Nextend:", style={"font-weight": "bold"}),
                            dcc.Input(
                                id=self.ids("nextend"),
                                debounce=True,
                                type="number",
                                required=True,
                                value=self.nextend,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ]
                    )
                ),
            ],
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            children=[
                html.Div(
                    style=self.set_grid_layout("1fr 1fr 1fr 1fr 1fr"),
                    children=[
                        self.well_layout,
                        self.well_options,
                        self.seismic_layout,
                        self.viz_options_layout,
                        html.Button(
                            id=self.ids("show_map"),
                            children="Show map",
                        ),
                    ],
                ),
                html.Div(
                    id=self.ids("viz_wrapper"),
                    style={"position": "relative"},
                    children=[
                        html.Div(
                            id=self.ids("map_wrapper"),
                            style={
                                "position": "absolute",
                                "width": "30%",
                                "height": "40%",
                                "right": 0,
                                "zIndex": 10000,
                                "visibility": "hidden",
                            },
                            children=LeafletMap(
                                id=self.ids("map"),
                                layers=[],
                                unitScale={},
                                autoScaleMap=True,
                                minZoom=-19,
                                updateMode="update",
                                mouseCoords={"position": "bottomright"},
                                colorBar={"position": "bottomleft"},
                            ),
                        ),
                        wcc.Graph(id=self.ids("graph")),
                    ],
                ),
            ],
        )

    @staticmethod
    def set_grid_layout(columns: str) -> Dict[str, str]:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.ids("graph"), "figure"),
            [
                Input(self.ids("wells"), "value"),
                Input(self.ids("cube"), "value"),
                Input(self.ids("options"), "value"),
                Input(self.ids("sampling"), "value"),
                Input(self.ids("nextend"), "value"),
            ],
        )
        def _render_section(
            wellfile: str,
            cubefile: str,
            options: List[str],
            sampling: Union[int, float, None],
            nextend: Union[int, float, None],
        ) -> dict:
            """Update cross section"""

            if sampling is None or nextend is None:
                raise PreventUpdate

            well = load_well(get_path(wellfile))
            xsect = XSectionFigure(
                well=well,
                zmin=self.zmin,
                zmax=self.zmax,
                nextend=int(nextend),
                sampling=int(sampling),
            )

            surfaces = [load_surface(get_path(surf)) for surf in self.surfacefiles]

            xsect.plot_surfaces(
                surfaces=surfaces,
                surfacenames=self.surfacenames,
                fill="show_surface_fill" in options,
            )

            if "show_seismic" in options:
                cube = load_cube_data(get_path(cubefile))
                xsect.plot_cube(cube)

            xsect.plot_well(
                zonelogname=self.zonelog if "show_zonelog" in options else None,
                zonemin=self.zonemin,
            )
            xsect.layout["margin"] = {"t": 0}
            return {"data": xsect.data, "layout": xsect.layout}

        @app.callback(
            [
                Output(self.ids("map_wrapper"), "style"),
                Output(self.ids("show_map"), "children"),
            ],
            [Input(self.ids("show_map"), "n_clicks")],
            [State(self.ids("map_wrapper"), "style")],
        )
        def _show_map(nclicks: int, style: dict) -> Tuple[dict, str]:
            """Show/hide map on button click"""
            btn = "Show Map"
            if not nclicks:
                raise PreventUpdate
            if nclicks % 2:
                style["visibility"] = "visible"
                btn = "Hide Map"
            else:
                style["visibility"] = "hidden"
            return style, btn

        @app.callback(
            Output(self.ids("map"), "layers"),
            [Input(self.ids("wells"), "value")],
        )
        def _render_surface(wellfile: str) -> List[Dict[str, Any]]:
            """Update map"""
            wellname = Path(wellfile).stem
            wellfile = get_path(wellfile)
            surface = load_surface(get_path(self.surfacefiles[0]))
            well = load_well(wellfile)
            s_layer = SurfaceLeafletModel(surface, name=self.surfacenames[0]).layer
            well_layer = make_well_layer(well, wellname)
            return [s_layer, well_layer]

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            *[(get_path, [{"path": fn}]) for fn in self.segyfiles],
            *[(get_path, [{"path": fn}]) for fn in self.surfacefiles],
            *[(get_path, [{"path": fn}]) for fn in self.wellfiles],
        ]


@webvizstore
def get_path(path: str) -> Path:
    return Path(path)



# pylint: skip-file
import logging
import time
from pathlib import Path

from .well_provider import WellProvider
from .well_provider_factory import WellProviderFactory


def main() -> None:
    print()
    print("## Running WellProvider experiments")
    print("## =================================================")

    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s %(levelname)-3s [%(name)s]: %(message)s",
    )
    logging.getLogger("webviz_subsurface").setLevel(level=logging.DEBUG)

    root_storage_dir = Path("/home/sigurdp/buf/webviz_storage_dir")

    well_folder = "../hk-webviz-subsurface-testdata/01_drogon_ahm/realization-0/iter-0/share/results/wells"
    well_suffix = ".rmswell"
    md_logname = None

    factory = WellProviderFactory(root_storage_dir, allow_storage_writes=True)

    provider: WellProvider = factory.create_from_well_files(
        well_folder=well_folder,
        well_suffix=well_suffix,
        md_logname=md_logname,
    )

    all_well_names = provider.well_names()
    print()
    print("all_well_names:")
    print("------------------------")
    print(*all_well_names, sep="\n")

    start_tim = time.perf_counter()

    for name in all_well_names:
        # w = provider.get_well_xtgeo_obj(name)
        wp = provider.get_well_path(name)

    elapsed_time_ms = int(1000 * (time.perf_counter() - start_tim))
    print(f"## get all wells took: {elapsed_time_ms}ms")

    well_name = "55_33-A-4"

    w = provider.get_well_xtgeo_obj(well_name)
    print(w.describe())
    print("w.mdlogname=", w.mdlogname)

    wp = provider.get_well_path(well_name)
    # print(wp)

    # comparewell = xtgeo.well_from_file(
    #     wfile=Path(well_folder) / "55_33-A-4.rmswell", mdlogname=md_logname
    # )
    # print(comparewell.describe())
    # print("comparewell.mdlogname=", comparewell.mdlogname)


# Running:
#   python -m webviz_subsurface._providers.well_provider.dev_experiments
# -------------------------------------------------------------------------
if __name__ == "__main__":
    main()



import datetime
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Sequence

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from pyarrow import feather

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._dataframe_utils import make_date_column_datetime_object
from ._field_metadata import create_vector_metadata_from_field_meta
from ._table_utils import (
    add_per_vector_min_max_to_table_schema_metadata,
    find_intersected_dates_between_realizations,
    find_min_max_for_numeric_table_columns,
    get_per_vector_min_max_from_schema_metadata,
)
from .ensemble_summary_provider import (
    EnsembleSummaryProvider,
    Frequency,
    VectorMetadata,
)

# Since PyArrow's actual compute functions are not seen by pylint
# pylint: disable=no-member


LOGGER = logging.getLogger(__name__)


def _create_float_downcasting_schema(schema: pa.Schema) -> pa.Schema:
    dt_float64 = pa.float64()
    dt_float32 = pa.float32()
    types = schema.types
    for idx, typ in enumerate(types):
        if typ == dt_float64:
            types[idx] = dt_float32

    field_list = zip(schema.names, types)
    return pa.schema(field_list)


def _set_date_column_type_to_timestamp_ms(schema: pa.Schema) -> pa.Schema:
    dt_timestamp_ms = pa.timestamp("ms")

    indexof_date_field = schema.get_field_index("DATE")

    types = schema.types
    types[indexof_date_field] = dt_timestamp_ms

    field_list = zip(schema.names, types)
    return pa.schema(field_list)


def _sort_table_on_date_then_real(table: pa.Table) -> pa.Table:
    indices = pc.sort_indices(
        table, sort_keys=[("DATE", "ascending"), ("REAL", "ascending")]
    )
    sorted_table = table.take(indices)
    return sorted_table


class ProviderImplArrowPresampled(EnsembleSummaryProvider):
    """Implements an EnsembleSummaryProvider without any resampling or interpolation"""

    def __init__(self, arrow_file_name: Path) -> None:
        self._arrow_file_name = str(arrow_file_name)

        LOGGER.debug(f"init with arrow file: {self._arrow_file_name}")
        timer = PerfTimer()

        source = pa.memory_map(self._arrow_file_name, "r")
        et_open_ms = timer.lap_ms()

        reader = pa.ipc.RecordBatchFileReader(source)
        et_create_reader_ms = timer.lap_ms()

        # Discover columns and realizations that are present in the file
        column_names_on_file = reader.schema.names
        self._vector_names: List[str] = [
            colname
            for colname in column_names_on_file
            if colname not in ["DATE", "REAL", "ENSEMBLE"]
        ]
        et_find_vec_names_ms = timer.lap_ms()

        unique_realizations_on_file = reader.read_all().column("REAL").unique()
        self._realizations: List[int] = unique_realizations_on_file.to_pylist()
        et_find_real_ms = timer.lap_ms()

        # We'll try and keep the file open for the life-span of the provider.
        # Done to try and stop blobfuse from throwing the file out of its cache.
        self._cached_reader = reader

        # For testing, uncomment code below and we will be more aggressive
        # and keep the "raw" table in memory
        self._cached_full_table = None
        # self._cached_full_table = reader.read_all()

        LOGGER.debug(
            f"init took: {timer.elapsed_s():.2f}s, "
            f"(open={et_open_ms}ms, create_reader={et_create_reader_ms}ms, "
            f"find_vec_names={et_find_vec_names_ms}ms, find_real={et_find_real_ms}ms), "
            f"#vector_names={len(self._vector_names)}, "
            f"#realization={len(self._realizations)}"
        )

        if not self._realizations:
            raise ValueError("Init from backing store failed NO realizations")
        if not self._vector_names:
            raise ValueError("Init from backing store failed NO vector_names")

    @staticmethod
    def write_backing_store_from_ensemble_dataframe(
        storage_dir: Path, storage_key: str, ensemble_df: pd.DataFrame
    ) -> None:
        @dataclass
        class Elapsed:
            convert_date_s: float = -1
            table_from_pandas_s: float = -1
            find_and_store_min_max_s: float = -1
            sorting_s: float = -1
            write_s: float = -1

        elapsed = Elapsed()

        arrow_file_name = storage_dir / (storage_key + ".arrow")
        LOGGER.debug(
            f"Writing backing store from ensemble dataframe to arrow file: {arrow_file_name}"
        )
        timer = PerfTimer()

        # Force data type in the incoming DataFrame's DATE column to datetime.datetime objects
        # This is the first step in coercing pyarrow to always store DATEs as timestamps
        ensemble_df = make_date_column_datetime_object(ensemble_df)
        elapsed.convert_date_s = timer.lap_s()

        # By default, we'll now end up with a schema that has timestamp[ns] for the DATE column
        # We therefore modify the retrieved schema and specify usage of timestamp[ms] instead
        default_schema = pa.Schema.from_pandas(ensemble_df, preserve_index=False)
        schema_to_use = _set_date_column_type_to_timestamp_ms(default_schema)

        # For experimenting with conversion to float
        # timer.lap_s()
        # schema_to_use = _create_float_downcasting_schema(schema_to_use)
        # LOGGER.info(
        #     f"Created schema for float downcasting in : {timer.lap_s():.2f}s"
        # )

        timer.lap_s()
        table = pa.Table.from_pandas(
            ensemble_df, schema=schema_to_use, preserve_index=False
        )
        elapsed.table_from_pandas_s = timer.lap_s()

        # We're done with the dataframe
        del ensemble_df

        # Find per column min/max values and then store them as metadata on table's schema
        timer.lap_ms()
        per_vector_min_max = find_min_max_for_numeric_table_columns(table)
        table = add_per_vector_min_max_to_table_schema_metadata(
            table, per_vector_min_max
        )
        elapsed.find_and_store_min_max_s = timer.lap_s()

        table = _sort_table_on_date_then_real(table)
        elapsed.sorting_s = timer.lap_s()

        # feather.write_feather(table, dest=arrow_file_name)
        feather.write_feather(table, dest=arrow_file_name, compression="uncompressed")
        elapsed.write_s = timer.lap_s()

        LOGGER.debug(
            f"Wrote backing store to arrow file in: {timer.elapsed_s():.2f}s ("
            f"convert_date={elapsed.convert_date_s:.2f}s, "
            f"table_from_pandas={elapsed.table_from_pandas_s:.2f}s, "
            f"find_and_store_min_max={elapsed.find_and_store_min_max_s:.2f}s, "
            f"sorting={elapsed.sorting_s:.2f}s, "
            f"write={elapsed.write_s:.2f}s)"
        )

    @staticmethod
    def write_backing_store_from_per_realization_tables(
        storage_dir: Path, storage_key: str, per_real_tables: Dict[int, pa.Table]
    ) -> None:
        @dataclass
        class Elapsed:
            concat_tables_s: float = -1
            build_add_real_col_s: float = -1
            sorting_s: float = -1
            find_and_store_min_max_s: float = -1
            write_s: float = -1

        elapsed = Elapsed()

        arrow_file_name = storage_dir / (storage_key + ".arrow")
        LOGGER.debug(
            f"Writing backing store from per real tables to arrow file: {arrow_file_name}"
        )
        timer = PerfTimer()

        unique_column_names = set()
        for table in per_real_tables.values():
            unique_column_names.update(table.schema.names)
        LOGGER.debug(
            f"Concatenating {len(per_real_tables)} tables with "
            f"{len(unique_column_names)} unique column names"
        )

        timer.lap_s()
        full_table = pa.concat_tables(per_real_tables.values(), promote=True)
        elapsed.concat_tables_s = timer.lap_s()

        real_arr = np.empty(full_table.num_rows, np.int32)
        table_start_idx = 0
        for real_num, real_table in per_real_tables.items():
            real_arr[table_start_idx : table_start_idx + real_table.num_rows] = real_num
            table_start_idx += real_table.num_rows

        full_table = full_table.add_column(0, "REAL", pa.array(real_arr))
        elapsed.build_add_real_col_s = timer.lap_s()

        # Find per column min/max values and then store them as metadata on table's schema
        per_vector_min_max = find_min_max_for_numeric_table_columns(full_table)
        full_table = add_per_vector_min_max_to_table_schema_metadata(
            full_table, per_vector_min_max
        )
        elapsed.find_and_store_min_max_s = timer.lap_s()

        full_table = _sort_table_on_date_then_real(full_table)
        elapsed.sorting_s = timer.lap_s()

        # feather.write_feather(full_table, dest=arrow_file_name)
        feather.write_feather(
            full_table, dest=arrow_file_name, compression="uncompressed"
        )
        elapsed.write_s = timer.lap_s()

        LOGGER.debug(
            f"Wrote backing store to arrow file in: {timer.elapsed_s():.2f}s ("
            f"concat_tables={elapsed.concat_tables_s:.2f}s, "
            f"build_add_real_col={elapsed.build_add_real_col_s:.2f}s, "
            f"sorting={elapsed.sorting_s:.2f}s, "
            f"find_and_store_min_max={elapsed.find_and_store_min_max_s:.2f}s, "
            f"write={elapsed.write_s:.2f}s)"
        )

    @staticmethod
    def from_backing_store(
        storage_dir: Path, storage_key: str
    ) -> Optional["ProviderImplArrowPresampled"]:

        arrow_file_name = storage_dir / (storage_key + ".arrow")
        if arrow_file_name.is_file():
            return ProviderImplArrowPresampled(arrow_file_name)

        return None

    def _get_or_read_schema(self) -> pa.Schema:
        if self._cached_full_table:
            return self._cached_full_table.schema
        if self._cached_reader:
            return self._cached_reader.schema

        source = pa.memory_map(self._arrow_file_name, "r")
        return pa.ipc.RecordBatchFileReader(source).schema

    def _get_or_read_table(self, columns: List[str]) -> pa.Table:
        if self._cached_full_table:
            return self._cached_full_table.select(columns)
        if self._cached_reader:
            return self._cached_reader.read_all().select(columns)

        source = pa.memory_map(self._arrow_file_name, "r")
        reader = pa.ipc.RecordBatchFileReader(source)
        return reader.read_all().select(columns)

    def vector_names(self) -> List[str]:
        return self._vector_names

    def vector_names_filtered_by_value(
        self,
        exclude_all_values_zero: bool = False,
        exclude_constant_values: bool = False,
    ) -> List[str]:

        timer = PerfTimer()

        schema = self._get_or_read_schema()
        et_read_ms = timer.lap_ms()

        per_vector_min_max = get_per_vector_min_max_from_schema_metadata(schema)
        et_get_min_max_ms = timer.lap_ms()

        ret_vec_names: List[str] = []
        for vec_name in self._vector_names:
            minval = per_vector_min_max[vec_name]["min"]
            maxval = per_vector_min_max[vec_name]["max"]

            if minval == maxval:
                if exclude_constant_values:
                    continue

                if exclude_all_values_zero and minval == 0:
                    continue

            ret_vec_names.append(vec_name)
        et_filter_ms = timer.lap_ms()

        LOGGER.debug(
            f"vector_names_filtered_by_value() took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"get_min_max={et_get_min_max_ms}ms, "
            f"filter={et_filter_ms}ms)"
        )

        return ret_vec_names

        # table = self._get_or_read_table(self._vector_names)
        # ret_vec_names: List[str] = []
        # for vec_name in self._vector_names:
        #     minmax = pc.min_max(table[vec_name])
        #     if minmax.get("min") == minmax.get("max"):
        #         if exclude_constant_values:
        #             continue

        #         if exclude_all_values_zero and minmax.get("min") == pa.scalar(0.0):
        #             continue

        #     ret_vec_names.append(vec_name)

        # return ret_vec_names

    def realizations(self) -> List[int]:
        return self._realizations

    def vector_metadata(self, vector_name: str) -> Optional[VectorMetadata]:
        schema = self._get_or_read_schema()
        field = schema.field(vector_name)
        return create_vector_metadata_from_field_meta(field)

    def supports_resampling(self) -> bool:
        return False

    def dates(
        self,
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> List[datetime.datetime]:

        if resampling_frequency is not None:
            raise ValueError("Resampling is not supported by this provider")

        timer = PerfTimer()

        table = self._get_or_read_table(["DATE", "REAL"])
        et_read_ms = timer.lap_ms()

        if realizations:
            mask = pc.is_in(table["REAL"], value_set=pa.array(realizations))
            table = table.filter(mask)
        et_filter_ms = timer.lap_ms()

        intersected_dates = find_intersected_dates_between_realizations(table)
        et_find_unique_ms = timer.lap_ms()

        LOGGER.debug(
            f"dates() took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"filter={et_filter_ms}ms, "
            f"find_unique={et_find_unique_ms}ms)"
        )

        return intersected_dates.astype(datetime.datetime).tolist()

    def get_vectors_df(
        self,
        vector_names: Sequence[str],
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:

        if resampling_frequency is not None:
            raise ValueError("Resampling is not supported by this provider")

        timer = PerfTimer()

        columns_to_get = ["DATE", "REAL"]
        columns_to_get.extend(vector_names)
        table = self._get_or_read_table(columns_to_get)
        et_read_ms = timer.lap_ms()

        if realizations is not None:
            mask = pc.is_in(table["REAL"], value_set=pa.array(realizations))
            table = table.filter(mask)
        et_filter_ms = timer.lap_ms()

        df = table.to_pandas(timestamp_as_object=True)
        # df = table.to_pandas(split_blocks=True, self_destruct=True)
        # del table  # not necessary, but a good practice
        et_to_pandas_ms = timer.lap_ms()

        LOGGER.debug(
            f"get_vectors_df() took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"filter={et_filter_ms}ms, "
            f"to_pandas={et_to_pandas_ms}ms), "
            f"#vecs={len(vector_names)}, "
            f"#real={len(realizations) if realizations is not None else 'all'}, "
            f"df.shape={df.shape}, file={Path(self._arrow_file_name).name}"
        )

        return df

    def get_vectors_for_date_df(
        self,
        date: datetime.datetime,
        vector_names: Sequence[str],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:

        timer = PerfTimer()

        columns_to_get = ["DATE", "REAL"]
        columns_to_get.extend(vector_names)
        table = self._get_or_read_table(columns_to_get)
        et_read_ms = timer.lap_ms()

        # Note that we use MS here to be aligned with storage type in arrow file
        lookup_date = pa.scalar(date, type=pa.timestamp("ms"))
        mask = pc.equal(table["DATE"], lookup_date)

        if realizations:
            real_mask = pc.is_in(table["REAL"], value_set=pa.array(realizations))
            mask = pc.and_(mask, real_mask)

        table = table.drop(["DATE"])

        # table = table.filter(mask).combine_chunks()
        table = table.filter(mask)
        et_filter_ms = timer.lap_ms()

        df = table.to_pandas()
        # df = table.to_pandas(split_blocks=True, zero_copy_only=True)
        # del table  # not necessary, but a good practice
        et_to_pandas_ms = timer.lap_ms()

        LOGGER.debug(
            f"get_vectors_for_date_df() took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"filter={et_filter_ms}ms, "
            f"to_pandas={et_to_pandas_ms}ms), "
            f"#vecs={len(vector_names)}, "
            f"#real={len(realizations) if realizations else 'all'}, "
            f"df.shape={df.shape}, file={Path(self._arrow_file_name).name}"
        )

        return df



from typing import Callable, Dict, List, Tuple, Union

from dash import MATCH, Dash, Input, Output


def update_intersection_source(
    app: Dash, get_uuid: Callable, surface_geometry: Dict
) -> None:
    @app.callback(
        Output(
            {"id": get_uuid("intersection-data"), "element": "well-wrapper"}, "style"
        ),
        Output(
            {"id": get_uuid("intersection-data"), "element": "xline-wrapper"}, "style"
        ),
        Output(
            {"id": get_uuid("intersection-data"), "element": "yline-wrapper"}, "style"
        ),
        Input({"id": get_uuid("intersection-data"), "element": "source"}, "value"),
    )
    def _show_source_details(source: str) -> Tuple[Dict, Dict, Dict]:
        hide = {"display": "none"}
        show = {"display": "inline"}
        if source == "well":
            return show, hide, hide
        if source == "xline":
            return hide, show, hide
        if source == "yline":
            return hide, hide, show
        return hide, hide, hide

    @app.callback(
        Output(
            {"id": get_uuid("map"), "element": "stored_xline"},
            "data",
        ),
        Input(
            {
                "id": get_uuid("intersection-data"),
                "cross-section": "xline",
                "element": "value",
            },
            "value",
        ),
    )
    def _store_xline(x_value: Union[float, int]) -> List:

        return [
            [x_value, surface_geometry["ymin"]],
            [x_value, surface_geometry["ymax"]],
        ]

    @app.callback(
        Output(
            {"id": get_uuid("map"), "element": "stored_yline"},
            "data",
        ),
        Input(
            {
                "id": get_uuid("intersection-data"),
                "cross-section": "yline",
                "element": "value",
            },
            "value",
        ),
    )
    def _store_yline(y_value: Union[float, int]) -> List:

        return [
            [surface_geometry["xmin"], y_value],
            [surface_geometry["xmax"], y_value],
        ]

    @app.callback(
        Output(
            {
                "id": get_uuid("intersection-data"),
                "cross-section": MATCH,
                "element": "value",
            },
            "step",
        ),
        Input(
            {
                "id": get_uuid("intersection-data"),
                "cross-section": MATCH,
                "element": "step",
            },
            "value",
        ),
    )
    def _set_cross_section_step(step: Union[float, int]) -> Union[float, int]:
        return step



from pathlib import Path
from typing import Dict, Optional, TypedDict

import yaml
from webviz_subsurface_components import VectorDefinition

# JSON Schema for user defined vector definitions
# Used as schema input for json_schema.validate()
USER_DEFINED_VECTOR_DEFINITIONS_JSON_SCHEMA = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "required": ["description"],
        "properties": {
            "description": {
                "type": "string",
                "maxLength": 50,
            },
            "type": {
                "type": "string",
                "maxLength": 15,
            },
        },
        "additionalProperties": False,
    },
}


class ConfigRequiredUserDefinedVectorDefinition(TypedDict):
    """Definition of required user defined vector definition data

    `Description:`
    Data type to represent required user defined data for a vector definition

    `Required keys`:
    * description: str - User defined description for vector
    """

    description: str


class ConfigUserDefinedVectorDefinition(
    ConfigRequiredUserDefinedVectorDefinition, total=False
):
    """Definition of user defined vector definition data

    Contains both required and non-required keys.

    `Description:`
    Data type to represent user defined data for a vector definition

    `Required keys`:
    * description: str - User defined description for vector

    `Non-required keys`:
    * type: str - Vector type used for VectorSelector category type
    """

    type: str


def create_user_defined_vector_descriptions_from_config(
    user_defined_vector_data_path: Optional[Path],
) -> Dict[str, VectorDefinition]:
    """Create user defined vector definitions from config

    `Input:`
    Path for yaml-file containing user defined vector definitions

    `Return:`
    Dict with vector as name, and user defined vector definition object as value.
    """
    output: Dict[str, VectorDefinition] = {}

    if user_defined_vector_data_path is None:
        return output

    vector_data_dict: Dict[str, ConfigUserDefinedVectorDefinition] = yaml.safe_load(
        user_defined_vector_data_path.read_text()
    )

    for vector, vector_data in vector_data_dict.items():
        _description = vector_data.get("description", "")
        _type = vector_data.get("type", "others")
        output[vector] = VectorDefinition(description=_description, type=_type)

    return output



from typing import Any, Callable, List, Optional, Tuple

import pandas as pd
import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import ALL, Input, Output, State, callback, callback_context, no_update
from dash.exceptions import PreventUpdate
from webviz_config import WebvizConfigTheme

from webviz_subsurface._components.tornado._tornado_bar_chart import TornadoBarChart
from webviz_subsurface._components.tornado._tornado_data import TornadoData
from webviz_subsurface._components.tornado._tornado_table import TornadoTable
from webviz_subsurface._figures import create_figure
from webviz_subsurface._models import InplaceVolumesModel

from ..utils.table_and_figure_utils import (
    create_data_table,
    create_table_columns,
    update_tornado_figures_xaxis,
)
from ..utils.utils import update_relevant_components
from ..views.tornado_view import tornado_error_layout, tornado_plots_layout


# pylint: disable=too-many-locals, too-many-statements
def tornado_controllers(
    get_uuid: Callable, volumemodel: InplaceVolumesModel, theme: WebvizConfigTheme
) -> None:
    @callback(
        Output({"id": get_uuid("main-tornado"), "page": ALL}, "children"),
        Input(get_uuid("selections"), "data"),
        State(get_uuid("page-selected"), "data"),
        State({"id": get_uuid("main-tornado"), "page": ALL}, "id"),
    )
    def _update_tornado_pages(
        selections: dict, page_selected: str, id_list: list
    ) -> go.Figure:

        if page_selected not in ["torn_multi", "torn_bulk_inplace"]:
            raise PreventUpdate

        selections = selections[page_selected]
        if not selections["update"]:
            raise PreventUpdate

        subplots = selections["Subplots"] is not None
        groups = ["REAL", "ENSEMBLE", "SENSNAME", "SENSCASE", "SENSTYPE"]
        if subplots and selections["Subplots"] not in groups:
            groups.append(selections["Subplots"])

        filters = selections["filters"].copy()

        figures = []
        tables = []
        responses = (
            ["BULK", selections["Response"]]
            if page_selected == "torn_bulk_inplace"
            else [selections["Response"]]
        )
        for response in responses:
            if not (response == "BULK" and page_selected == "torn_bulk_inplace"):
                if selections["Reference"] not in selections["Sensitivities"]:
                    selections["Sensitivities"].append(selections["Reference"])
                filters.update(SENSNAME=selections["Sensitivities"])

            dframe = volumemodel.get_df(filters=filters, groups=groups)

            if not dframe.empty:
                dframe.rename(columns={response: "VALUE"}, inplace=True)
                df_groups = (
                    dframe.groupby(selections["Subplots"])
                    if subplots
                    else [(None, dframe)]
                )
                for group, df in df_groups:
                    if selections["Reference"] in df["SENSNAME"].unique():
                        tornado_data = TornadoData(
                            dframe=df,
                            reference=selections["Reference"],
                            response_name=response,
                            scale=selections["Scale"],
                            cutbyref=bool(selections["Remove no impact"]),
                        )
                        figure, table_data, columns = tornado_figure_and_table(
                            tornado_data=tornado_data,
                            response=response,
                            selections=selections,
                            theme=theme,
                            sensitivity_colors=sens_colors(),
                            font_size=max((20 - (0.4 * len(df_groups))), 10),
                            group=group,
                            use_si_format=response in volumemodel.volume_columns,
                        )
                        figures.append(figure)
                        tables.append(table_data)

                        if (
                            response == selections["Response"]
                            and selections["bottom_viz"] == "realplot"
                            and not subplots
                        ):
                            realplot = create_realplot(
                                df=tornado_data.real_df,
                                sensitivity_colors=sens_colors(),
                            )

        if selections["Shared axis"] and selections["Scale"] != "True":
            update_tornado_figures_xaxis(figures)

        bottom_display: list = []
        if selections["bottom_viz"] == "table" and figures:
            bottom_display = create_data_table(
                columns=columns,
                selectors=[selections["Subplots"]] if subplots else [],
                data=[x for table in tables for x in table],
                height="39vh",
                table_id={"table_id": f"{page_selected}-torntable"},
            )
        elif selections["bottom_viz"] == "realplot" and figures:
            bottom_display = [
                wcc.Graph(
                    config={"displayModeBar": False},
                    style={"height": "40vh"},
                    figure=realplot,
                )
                if not subplots
                else "Realization plot not available when `Subplots` is active"
            ]

        return update_relevant_components(
            id_list=id_list,
            update_info=[
                {
                    "new_value": tornado_plots_layout(
                        figures=figures, bottom_display=bottom_display
                    )
                    if figures
                    else tornado_error_layout(
                        "No data left after filtering"
                        if dframe.empty
                        else f"Reference sensitivity '{selections['Reference']}' not in input data"
                    ),
                    "conditions": {"page": page_selected},
                }
            ],
        )

    @callback(
        Output(
            {"id": get_uuid("selections"), "selector": ALL, "tab": "tornado"}, "options"
        ),
        Output(
            {"id": get_uuid("selections"), "selector": ALL, "tab": "tornado"}, "value"
        ),
        Output(
            {"id": get_uuid("selections"), "selector": ALL, "tab": "tornado"},
            "disabled",
        ),
        Input(get_uuid("page-selected"), "data"),
        State(get_uuid("tabs"), "value"),
        State({"id": get_uuid("selections"), "selector": ALL, "tab": "tornado"}, "id"),
        State(
            {"id": get_uuid("selections"), "selector": ALL, "tab": "tornado"}, "value"
        ),
        State(get_uuid("selections"), "data"),
    )
    def _update_tornado_selections(
        selected_page: str,
        selected_tab: str,
        selector_ids: list,
        selector_values: list,
        previous_selection: Optional[dict],
    ) -> tuple:
        if selected_tab != "tornado" or previous_selection is None:
            raise PreventUpdate

        ctx = callback_context.triggered[0]
        initial_page_load = selected_page not in previous_selection

        selections: Any = (
            previous_selection.get(selected_page)
            if "page-selected" in ctx["prop_id"] and selected_page in previous_selection
            else {
                id_value["selector"]: values
                for id_value, values in zip(selector_ids, selector_values)
            }
        )

        settings = {}
        if selected_page == "torn_bulk_inplace":
            volume_options = [
                x for x in ["STOIIP", "GIIP"] if x in volumemodel.responses
            ]
            settings["Response"] = {
                "options": [{"label": i, "value": i} for i in volume_options],
                "value": volume_options[0]
                if initial_page_load
                else selections["Response"],
                "disabled": len(volume_options) == 1,
            }
        else:
            responses = [x for x in volumemodel.responses if x not in ["BO", "BG"]]
            settings["Response"] = {
                "options": [{"label": i, "value": i} for i in responses],
                "disabled": False,
                "value": selections["Response"],
            }

        settings["bottom_viz"] = {"value": selections["bottom_viz"]}

        disable_subplots = selected_page != "torn_multi"
        settings["Subplots"] = {
            "disabled": disable_subplots,
            "value": None if disable_subplots else selections["Subplots"],
        }

        return tuple(
            update_relevant_components(
                id_list=selector_ids,
                update_info=[
                    {
                        "new_value": values.get(prop, no_update),
                        "conditions": {"selector": selector},
                    }
                    for selector, values in settings.items()
                ],
            )
            for prop in ["options", "value", "disabled"]
        )

    def sens_colors() -> dict:
        colors = [
            "#FF1243",
            "#243746",
            "#007079",
            "#80B7BC",
            "#919BA2",
            "#BE8091",
            "#B2D4D7",
            "#FF597B",
            "#BDC3C7",
            "#D8B2BD",
            "#FFE7D6",
            "#D5EAF4",
            "#FF88A1",
        ]
        sensitivities = volumemodel.dataframe["SENSNAME"].unique()
        return dict(zip(sensitivities, colors * 10))


def tornado_figure_and_table(
    tornado_data: TornadoData,
    response: str,
    selections: dict,
    theme: WebvizConfigTheme,
    sensitivity_colors: dict,
    font_size: float,
    use_si_format: bool,
    group: Optional[str] = None,
) -> Tuple[go.Figure, List[dict], List[dict]]:

    figure = TornadoBarChart(
        tornado_data=tornado_data,
        plotly_theme=theme.plotly_theme,
        label_options=selections["labeloptions"],
        number_format="#.3g",
        locked_si_prefix=None if use_si_format else "",
        use_true_base=selections["Scale"] == "True",
        show_realization_points=bool(selections["real_scatter"]),
        show_reference=selections["torn_ref"],
        color_by_sensitivity=selections["color_by_sens"],
        sensitivity_color_map=sensitivity_colors,
    ).figure

    figure.update_xaxes(side="bottom", title=None).update_layout(
        title_text=f"Tornadoplot for {response} <br>"
        + f"Fluid zone: {(' + ').join(selections['filters']['FLUID_ZONE'])}"
        if group is None
        else f"{response} {group}",
        title_font_size=font_size,
        margin={"t": 70},
    )

    table_data, columns = create_tornado_table(
        tornado_data,
        subplots=selections["Subplots"],
        group=group,
        use_si_format=use_si_format,
    )
    return figure, table_data, columns


def create_realplot(df: pd.DataFrame, sensitivity_colors: dict) -> go.Figure:
    senscasecolors = {
        senscase: sensitivity_colors[sensname]
        for senscase, sensname in zip(df["sensname_case"], df["sensname"])
    }

    return (
        create_figure(
            plot_type="bar",
            data_frame=df,
            x="REAL",
            y="VALUE",
            color="sensname_case",
            color_discrete_map=senscasecolors,
            barmode="overlay",
            custom_data=["casetype"],
            yaxis={"range": [df["VALUE"].min() * 0.7, df["VALUE"].max() * 1.1]},
            opacity=0.85,
        )
        .update_layout(legend={"orientation": "h", "yanchor": "bottom", "y": 1.02})
        .update_layout(legend_title_text="")
        .for_each_trace(
            lambda t: (
                t.update(marker_line_color="black")
                if t["customdata"][0][0] == "high"
                else t.update(marker_line_color="white", marker_line_width=2)
            )
            if t["customdata"][0][0] != "mc"
            else None
        )
    )


def create_tornado_table(
    tornado_data: TornadoData,
    subplots: str,
    group: Optional[str],
    use_si_format: bool,
) -> Tuple[List[dict], List[dict]]:

    tornado_table = TornadoTable(
        tornado_data=tornado_data,
        use_si_format=use_si_format,
        precision=4 if use_si_format else 3,
    )
    table_data = tornado_table.as_plotly_table
    for data in table_data:
        data["Reference"] = tornado_data.reference_average
        if group is not None:
            data[subplots] = group

    columns = create_table_columns(columns=[subplots]) if subplots is not None else []
    columns.extend(tornado_table.columns)
    columns.extend(
        create_table_columns(
            columns=["Reference"],
            use_si_format=["Reference"] if use_si_format else [],
        )
    )
    return table_data, columns



import json
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import xtgeo
from dash import ClientsideFunction, Dash, Input, Output, State, callback_context
from dash.exceptions import PreventUpdate

from webviz_subsurface._components import ColorPicker
from webviz_subsurface._models import SurfaceSetModel, WellSetModel

from ..figures.intersection import (
    get_plotly_trace_realization_surface,
    get_plotly_trace_statistical_surface,
    get_plotly_trace_well_trajectory,
    get_plotly_traces_uncertainty_envelope,
    get_plotly_zonelog_trace,
)


# pylint: disable=too-many-statements
def update_intersection(
    app: Dash,
    get_uuid: Callable,
    surface_set_models: Dict[str, SurfaceSetModel],
    well_set_model: WellSetModel,
    color_picker: ColorPicker,
    zonelog: Optional[str] = None,
) -> None:
    @app.callback(
        Output(get_uuid("intersection-graph-data"), "data"),
        Input(get_uuid("apply-intersection-data-selections"), "n_clicks"),
        Input(
            {"id": get_uuid("intersection-data"), "element": "source"},
            "value",
        ),
        Input({"id": get_uuid("map"), "element": "stored_polyline"}, "data"),
        Input({"id": get_uuid("map"), "element": "stored_xline"}, "data"),
        Input({"id": get_uuid("map"), "element": "stored_yline"}, "data"),
        Input({"id": get_uuid("intersection-data"), "element": "well"}, "value"),
        Input(get_uuid("realization-store"), "data"),
        State(
            {"id": get_uuid("intersection-data"), "element": "surface_attribute"},
            "value",
        ),
        State(
            {"id": get_uuid("intersection-data"), "element": "surface_names"}, "value"
        ),
        State(
            {"id": get_uuid("intersection-data"), "element": "calculation"},
            "value",
        ),
        State({"id": get_uuid("intersection-data"), "element": "ensembles"}, "value"),
        State({"id": get_uuid("intersection-data"), "element": "resolution"}, "value"),
        State({"id": get_uuid("intersection-data"), "element": "extension"}, "value"),
        State(color_picker.color_store_id, "data"),
    )
    # pylint: disable=too-many-arguments: disable=too-many-branches, too-many-locals
    def _store_intersection_traces(
        _apply_click: Optional[int],
        intersection_source: str,
        polyline: Optional[List],
        xline: Optional[List],
        yline: Optional[List],
        wellname: str,
        realizations: List[int],
        surfaceattribute: str,
        surfacenames: List[str],
        statistics: List[str],
        ensembles: str,
        resolution: float,
        extension: int,
        color_list: List[str],
    ) -> List:
        """Generate plotly traces for intersection figure and store clientside"""

        # TODO(Sigurd) Can we prohibit clearing of the sampling and extension input
        # fields (dcc.Input) in the client? Until we can, we must guard against sampling
        # and extension being None. This happens when the user clears the input field and we
        # have not yet found a solution that prohibits the input field from being cleared.
        # The situation can be slightly remedied by setting required=True which will highlight
        # the missing value with a red rectangle.
        if any(val is None for val in [resolution, extension]):
            raise PreventUpdate
        traces = []

        if intersection_source == "polyline":
            if polyline is None:
                return []
            fence_spec = get_fencespec_from_polyline(
                polyline, distance=resolution, atleast=5, nextend=extension / resolution
            )
        elif intersection_source == "xline":
            if xline is None:
                return []
            fence_spec = get_fencespec_from_polyline(
                xline, distance=resolution, atleast=5, nextend=extension / resolution
            )
        elif intersection_source == "yline":
            if yline is None:
                return []
            fence_spec = get_fencespec_from_polyline(
                yline, distance=resolution, atleast=5, nextend=extension / resolution
            )
        else:
            fence_spec = well_set_model.get_fence(
                well_name=wellname,
                distance=resolution,
                atleast=5,
                nextend=extension / resolution,
            )

        realizations = [int(real) for real in realizations]
        for ensemble in ensembles:
            surfset = surface_set_models[ensemble]
            for surfacename in surfacenames:
                color = color_picker.get_color(
                    color_list=color_list,
                    filter_query={
                        "surfacename": surfacename,
                        "ensemble": ensemble,
                    },
                )
                showlegend = True

                if statistics is not None:
                    for stat in ["Mean", "Min", "Max"]:
                        if stat in statistics:
                            trace = get_plotly_trace_statistical_surface(
                                surfaceset=surfset,
                                fence_spec=fence_spec,
                                calculation=stat,
                                legendname=f"{surfacename}({ensemble})",
                                name=surfacename,
                                attribute=surfaceattribute,
                                realizations=realizations,
                                showlegend=showlegend,
                                color=color,
                            )
                            traces.append(trace)
                            showlegend = False
                    if "Uncertainty envelope" in statistics:
                        envelope_traces = get_plotly_traces_uncertainty_envelope(
                            surfaceset=surfset,
                            fence_spec=fence_spec,
                            legendname=f"{surfacename}({ensemble})",
                            name=surfacename,
                            attribute=surfaceattribute,
                            realizations=realizations,
                            showlegend=showlegend,
                            color=color,
                        )
                        traces.extend(envelope_traces)
                        showlegend = False
                    if "Realizations" in statistics:
                        for real in realizations:
                            trace = get_plotly_trace_realization_surface(
                                surfaceset=surfset,
                                fence_spec=fence_spec,
                                legendname=f"{surfacename}({ensemble})",
                                name=surfacename,
                                attribute=surfaceattribute,
                                realization=real,
                                color=color,
                                showlegend=showlegend,
                            )
                            traces.append(trace)
                            showlegend = False
        if intersection_source == "well":
            well = well_set_model.get_well(wellname)
            traces.append(get_plotly_trace_well_trajectory(well))
            if well.zonelogname is not None:
                traces.extend(get_plotly_zonelog_trace(well, zonelog))

        return traces

    @app.callback(
        Output(get_uuid("intersection-graph-layout"), "data"),
        Input(get_uuid("intersection-graph-data"), "data"),
        Input(get_uuid("initial-intersection-graph-layout"), "data"),
        Input(
            {"id": get_uuid("intersection-data"), "element": "source"},
            "value",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "settings": "zrange_locks"},
            "value",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "settings": "zrange_min"},
            "value",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "settings": "zrange_max"},
            "value",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "settings": "ui_options"},
            "value",
        ),
        State(get_uuid("leaflet-map1"), "polyline_points"),
        State({"id": get_uuid("intersection-data"), "element": "well"}, "value"),
    )
    # pylint: disable=too-many-arguments, too-many-branches
    def _store_intersection_layout(
        data: List,
        initial_layout: Optional[dict],
        intersection_source: str,
        zrange_locks: str,
        zmin: Optional[float],
        zmax: Optional[float],
        ui_options: List[str],
        polyline: Optional[List],
        wellname: str,
    ) -> Dict:
        """Store intersection layout configuration clientside"""
        ctx = callback_context.triggered[0]
        if "ui_options" in ctx["prop_id"]:
            raise PreventUpdate

        # Set default layout
        layout: Dict = {
            "hovermode": "closest",
            "yaxis": {
                "autorange": "reversed",
                "showgrid": False,
                "zeroline": False,
                "title": "True vertical depth",
            },
            "xaxis": {
                "showgrid": False,
                "zeroline": False,
                "title": "Lateral resolution",
            },
            "plot_bgcolor": "rgba(0, 0, 0, 0)",
            "paper_bgcolor": "rgba(0, 0, 0, 0)",
        }

        # Update title to reflect source of cross-section calculation
        annotation_title = ["A", "A'"]
        if intersection_source in ["polyline", "xline", "yline"]:
            layout.update(
                {
                    "title": f"Intersection along {intersection_source} shown in Surface A"
                }
            )
            layout.get("xaxis", {}).update({"autorange": True})
            annotation_title = ["B", "B'"]
        if intersection_source == "well":
            layout["title"] = f"Intersection along well: {wellname}"

        # Set A-B annotations on plot
        layout["annotations"] = [
            {
                "x": 0,
                "y": 1,
                "xref": "paper",
                "yref": "paper",
                "text": f"<b>{annotation_title[0]}</b>",
                "font": {"size": 40},
                "showarrow": False,
            },
            {
                "x": 1,
                "y": 1,
                "xref": "paper",
                "yref": "paper",
                "text": f"<b>{annotation_title[1]}</b>",
                "font": {"size": 40},
                "showarrow": False,
            },
        ]
        # Update layout with any values provided from yaml configuration
        if initial_layout is not None:
            layout.update(initial_layout)

        # Return emptly plot layout if surface is source but no polyline is drawn
        if intersection_source == "polyline" and polyline is None:
            layout.update(
                {
                    "title": "Draw a random line from the toolbar on Surface A",
                }
            )
            return layout

        # Add any interactivily set range options
        if ui_options:
            if "uirevision" in ui_options:
                layout.update({"uirevision": "keep"})

        user_range = []
        if not (zmax is None and zmin is None):
            if "lock" in zrange_locks:
                if zmax is None:
                    zmax = max(
                        max(x for x in item["y"] if x is not None) for item in data
                    )
                if zmin is None:
                    zmin = min(
                        min(x for x in item["y"] if x is not None) for item in data
                    )
                user_range = [zmax, zmin]

            if "truncate" in zrange_locks:
                zmin_data = min(
                    min(x for x in item["y"] if x is not None) for item in data
                )
                zmax_data = max(
                    max(x for x in item["y"] if x is not None) for item in data
                )
                zmax = zmax if zmax is not None else zmax_data
                zmin = zmin if zmin is not None else zmin_data

                user_range = [min(zmax, zmax_data), max(zmin, zmin_data)]

        # Set y-axis range from depth range input if specified
        if user_range:
            layout.get("yaxis", {}).update({"autorange": False})
            layout.get("yaxis", {}).update(range=user_range)
        # Else autocalculate range if not intersecting a well
        elif intersection_source != "well":
            if "range" in layout.get("yaxis", {}):
                del layout["yaxis"]["range"]
            layout.get("yaxis", {}).update({"autorange": "reversed"})

        # Remove xaxis zero line
        layout.get("xaxis", {}).update({"zeroline": False, "showline": False})
        return layout

    # Store intersection data and layout to the plotly figure
    # Done clientside for performance
    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="set_dcc_figure"),
        Output(get_uuid("intersection-graph"), "figure"),
        Input(get_uuid("intersection-graph-layout"), "data"),
        State(get_uuid("intersection-graph-data"), "data"),
    )

    @app.callback(
        Output(
            {"id": get_uuid("intersection-data"), "settings": "zrange_min"},
            "max",
        ),
        Output(
            {"id": get_uuid("intersection-data"), "settings": "zrange_max"},
            "min",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "settings": "zrange_min"},
            "value",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "settings": "zrange_max"},
            "value",
        ),
    )
    def _set_min_max_for_range_input(
        zmin: Optional[float],
        zmax: Optional[float],
    ) -> Tuple[Optional[float], Optional[float]]:
        ctx = callback_context.triggered[0]
        if ctx["prop_id"] == ".":
            raise PreventUpdate

        return zmax, zmin

    @app.callback(
        Output(get_uuid("apply-intersection-data-selections"), "style"),
        Output(
            {
                "id": get_uuid("intersection-data"),
                "element": "stored_manual_update_options",
            },
            "data",
        ),
        Input(get_uuid("apply-intersection-data-selections"), "n_clicks"),
        Input(
            {"id": get_uuid("intersection-data"), "element": "surface_attribute"},
            "value",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "element": "surface_names"}, "value"
        ),
        Input(
            {"id": get_uuid("intersection-data"), "element": "calculation"},
            "value",
        ),
        Input({"id": get_uuid("intersection-data"), "element": "ensembles"}, "value"),
        Input({"id": get_uuid("intersection-data"), "element": "resolution"}, "value"),
        Input({"id": get_uuid("intersection-data"), "element": "extension"}, "value"),
        Input(color_picker.color_store_id, "data"),
        State(
            {
                "id": get_uuid("intersection-data"),
                "element": "stored_manual_update_options",
            },
            "data",
        ),
    )
    def _update_apply_button(
        _apply_click: Optional[int],
        surfaceattribute: str,
        surfacenames: List[str],
        statistics: List[str],
        ensembles: str,
        resolution: float,
        extension: int,
        color_list: List[str],
        previous_settings: Dict,
    ) -> Tuple[Dict, Dict]:

        ctx = callback_context.triggered[0]

        new_settings = {
            "surface_attribute": surfaceattribute,
            "surface_names": surfacenames,
            "calculation": statistics,
            "ensembles": ensembles,
            "resolution": resolution,
            "extension": extension,
            "colors": color_list,
        }
        # store selected settings if initial callback or apply button is pressed
        if (
            "apply-intersection-data-selections" in ctx["prop_id"]
            or ctx["prop_id"] == "."
        ):
            return {"background-color": "#E8E8E8"}, new_settings

        element = (
            "colors"
            if "colorpicker" in ctx["prop_id"]
            else json.loads(ctx["prop_id"].replace(".value", "")).get("element")
        )
        if new_settings[element] != previous_settings[element]:
            return {"background-color": "#7393B3", "color": "#fff"}, previous_settings
        return {"background-color": "#E8E8E8"}, previous_settings


def get_fencespec_from_polyline(
    coords: List, distance: float, atleast: int, nextend: Union[float, int]
) -> np.ndarray:
    """Create a fence specification from polyline coordinates"""
    poly = xtgeo.Polygons()
    poly.dataframe = pd.DataFrame(
        [
            {
                "X_UTME": c[0],
                "Y_UTMN": c[1],
                "Z_TVDSS": 0,
                "POLY_ID": 1,
                "NAME": "polyline",
            }
            for c in coords
        ]
    )
    return poly.get_fence(
        distance=distance, atleast=atleast, nextend=nextend, asnumpy=True
    )



from typing import Any, Dict, List

import plotly.graph_objects as go

from ......_utils.colors import find_intermediate_color, rgba_to_str
from ...._types import VfpParam
from ...._utils import VfpTable

RED = rgba_to_str((255, 18, 67, 1))
MID_COLOR = rgba_to_str((31, 119, 180, 1))
GREEN = rgba_to_str((62, 208, 62, 1))


class VfpFigureBuilder:
    """Class to build the VFP graph object"""

    def __init__(self, vfp_name: str) -> None:

        self._traces: List[Dict[str, Any]] = []
        self._layout = {
            "legend": {"orientation": "h"},
            "hovermode": "closest",
            "title": f"VFP table: {vfp_name}",
            "plot_bgcolor": "white",
        }

    def get_figure(self) -> Dict[str, Any]:
        return go.Figure(data=self._traces, layout=self._layout)

    def set_yaxis_settings(self, title: str, showgrid: bool = True) -> None:
        self._layout["yaxis"] = {"title": title, "showgrid": showgrid}

    def set_xaxis_settings(self, title: str, showgrid: bool = True) -> None:
        self._layout["xaxis"] = {"title": title, "showgrid": showgrid}

    def add_vfp_curve(
        self,
        rates: List[float],
        bhp_values: List[float],
        cmax: float,
        cmin: float,
        vfp_table: VfpTable,
        indices: Dict[VfpParam, int],
        color_by: VfpParam,
        showscale: bool,
    ) -> None:
        # pylint: disable=too-many-arguments
        """Adds one vfp trace to the vfp graph, formatted with
        color and hoverdata.

        Plotly apparently doesn't support colorscale on both markers and
        lines: https://github.com/plotly/plotly.js/issues/5259. So I needed
        to interpolate the line color manually in order to match the marker
        color.
        """
        cvalue = vfp_table.params[color_by][indices[color_by]]
        hovertext = "<br>".join(
            [
                f"{vfp_table.param_types[tpe].name}={vfp_table.params[tpe][idx]}"
                for tpe, idx in indices.items()
            ]
        )

        self._traces.append(
            {
                "x": rates,
                "y": bhp_values,
                "hovertext": hovertext,
                "hoverinfo": "y+x+text",
                "showlegend": False,
                "mode": "markers+lines",
                "marker": dict(
                    cmax=cmax,
                    cmin=cmin,
                    cmid=(cmin + cmax) / 2,
                    color=[cvalue] * len(rates),
                    colorscale=[[0, RED], [0.5, MID_COLOR], [1, GREEN]],
                    showscale=showscale,
                    colorbar={"title": vfp_table.param_types[color_by].value},
                ),
                "line": {"color": _get_color(cmax, cmin, cvalue)},
            }
        )


def _get_color(cmax: float, cmin: float, cvalue: float) -> str:
    """Returns an interpolated color"""
    if cmax < cmin:
        raise ValueError("'cmax' must be equal to or larger than 'cmin'")
    if cvalue < cmin or cvalue > cmax:
        raise ValueError("'cvalue' must be between 'cmin' and 'cmax'")
    if cmax == cmin:
        return MID_COLOR
    factor = (cvalue - cmin) / (cmax - cmin)
    if factor >= 0.5:
        return find_intermediate_color(MID_COLOR, GREEN, (factor - 0.5) / 0.5)
    return find_intermediate_color(RED, MID_COLOR, (factor) / 0.5)



import datetime
import json
from pathlib import Path
from typing import Callable, List, Tuple, Union
from uuid import uuid4

import numpy as np
import pandas as pd
import webviz_core_components as wcc
from dash import Dash, Input, Output, State, callback_context, dash_table, dcc, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._components import TornadoWidget
from webviz_subsurface._models import (
    EnsembleSetModel,
    caching_ensemble_set_model_factory,
)

from .._abbreviations.number_formatting import table_statistics_base
from .._abbreviations.reservoir_simulation import (
    historical_vector,
    simulation_unit_reformat,
    simulation_vector_description,
)
from .._datainput.fmu_input import find_sens_type, get_realizations
from .._utils.simulation_timeseries import (
    get_simulation_line_shape,
    set_simulation_line_shape_fallback,
)


# pylint: disable=too-many-instance-attributes
class ReservoirSimulationTimeSeriesOneByOne(WebvizPluginABC):
    """Visualizes reservoir simulation time series data for sensitivity studies based \
on a design matrix.

A tornado plot can be calculated interactively for each date/vector by selecting a date.
After selecting a date individual sensitivities can be selected to highlight the realizations
run with that sensitivity.

---
**Two main options for input data: Aggregated and read from UNSMRY.**

**Using aggregated data**
* **`csvfile_smry`:** Aggregated `csv` file for volumes with `REAL`, `ENSEMBLE`, `DATE` and \
    vector columns (absolute path or relative to config file).
* **`csvfile_parameters`:** Aggregated `csv` file for sensitivity information with `REAL`, \
    `ENSEMBLE`, `SENSNAME` and `SENSCASE` columns (absolute path or relative to config file).

**Using simulation time series data directly from `UNSMRY` files**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`column_keys`:** List of vectors to extract. If not given, all vectors \
    from the simulations will be extracted. Wild card asterisk `*` can be used.
* **`sampling`:** Time separation between extracted values. Can be e.g. `monthly` (default) or \
    `yearly`.

**Common optional settings for both input options**
* **`initial_vector`:** Initial vector to display
* **`line_shape_fallback`:** Fallback interpolation method between points. Vectors identified as \
    rates or phase ratios are always backfilled, vectors identified as cumulative (totals) are \
    always linearly interpolated. The rest use the fallback.
    Supported options:
    * `linear` (default)
    * `backfilled`
    * `hv`, `vh`, `hvh`, `vhv` and `spline` (regular Plotly options).

---
!> It is **strongly recommended** to keep the data frequency to a regular frequency (like \
`monthly` or `yearly`). This applies to both csv input and when reading from `UNSMRY` \
(controlled by the `sampling` key). This is because the statistics and fancharts are calculated \
per DATE over all realizations in an ensemble, and the available dates should therefore not \
differ between individual realizations of an ensemble.


**Using aggregated data**

* [Example of csvfile_smry]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/aggregated_data/smry.csv).

* [Example of csvfile_parameters]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/aggregated_data/parameters.csv).


**Using simulation time series data directly from `.UNSMRY` files**

Time series data are extracted automatically from the `UNSMRY` files in the individual
realizations, using the `fmu-ensemble` library. The `SENSNAME` and `SENSCASE` values are read
directly from the `parameters.txt` files of the individual realizations, assuming that these
exist. If the `SENSCASE` of a realization is `p10_p90`, the sensitivity case is regarded as a
**Monte Carlo** style sensitivity, otherwise the case is evaluated as a **scalar** sensitivity.

?> Using the `UNSMRY` method will also extract metadata like units, and whether the vector is a \
rate, a cumulative, or historical. Units are e.g. added to the plot titles, while rates and \
cumulatives are used to decide the line shapes in the plot. Aggregated data may on the other \
speed up the build of the app, as processing of `UNSMRY` files can be slow for large models.

!> The `UNSMRY` files are auto-detected by `fmu-ensemble` in the `eclipse/model` folder of the \
individual realizations. You should therefore not have more than one `UNSMRY` file in this \
folder, to avoid risk of not extracting the right data.
"""

    ENSEMBLE_COLUMNS = [
        "REAL",
        "ENSEMBLE",
        "DATE",
        "SENSCASE",
        "SENSNAME",
        "SENSTYPE",
        "RUNPATH",
    ]

    TABLE_STAT: List[Tuple[str, dict]] = [
        ("Sensitivity", {}),
        ("Case", {}),
    ] + table_statistics_base()

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csvfile_smry: Path = None,
        csvfile_parameters: Path = None,
        ensembles: list = None,
        column_keys: list = None,
        initial_vector: str = None,
        sampling: str = "monthly",
        line_shape_fallback: str = "linear",
    ) -> None:

        super().__init__()

        self.time_index = sampling
        self.column_keys = column_keys
        self.csvfile_smry = csvfile_smry
        self.csvfile_parameters = csvfile_parameters

        if csvfile_smry and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_smry" and "csvfile_parameters" or '
                '"ensembles"'
            )
        if csvfile_smry and csvfile_parameters:
            self.smry = read_csv(csvfile_smry)
            self.parameters = read_csv(csvfile_parameters)
            self.parameters["SENSTYPE"] = self.parameters.apply(
                lambda row: find_sens_type(row.SENSCASE), axis=1
            )
            self.smry_meta = None

        elif ensembles:
            self.ens_paths = {
                ensemble: webviz_settings.shared_settings["scratch_ensembles"][ensemble]
                for ensemble in ensembles
            }
            self.emodel: EnsembleSetModel = (
                caching_ensemble_set_model_factory.get_or_create_model(
                    ensemble_paths=self.ens_paths,
                    time_index=self.time_index,
                    column_keys=self.column_keys,
                )
            )
            self.smry = self.emodel.get_or_load_smry_cached()
            self.smry_meta = self.emodel.load_smry_meta()

            # Extract realizations and sensitivity information
            self.parameters = get_realizations(
                ensemble_paths=self.ens_paths, ensemble_set_name="EnsembleSet"
            )
        else:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_smry" and "csvfile_parameters" or '
                '"ensembles"'
            )
        self.smry_cols = [
            c
            for c in self.smry.columns
            if c not in ReservoirSimulationTimeSeriesOneByOne.ENSEMBLE_COLUMNS
            and historical_vector(c, self.smry_meta, False) not in self.smry.columns
        ]
        self.initial_vector = (
            initial_vector
            if initial_vector and initial_vector in self.smry_cols
            else self.smry_cols[0]
        )
        self.ensembles = list(self.parameters["ENSEMBLE"].unique())
        self.line_shape_fallback = set_simulation_line_shape_fallback(
            line_shape_fallback
        )
        self.tornadoplot = TornadoWidget(
            app, webviz_settings, self.parameters, allow_click=True
        )
        self.uid = uuid4()
        self.theme = webviz_settings.theme
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.ids("layout"),
                "content": (
                    "Dashboard displaying time series from a sensitivity study."
                ),
            },
            {
                "id": self.ids("graph-wrapper"),
                "content": (
                    "Selected time series displayed per realization. "
                    "Click in the plot to calculate tornadoplot for the "
                    "corresponding date, then click on the tornado plot to "
                    "highlight the corresponding sensitivity."
                ),
            },
            {
                "id": self.ids("table"),
                "content": (
                    "Table statistics for all sensitivities for the selected date."
                ),
            },
            *self.tornadoplot.tour_steps,
            {"id": self.ids("vector"), "content": "Select time series"},
            {"id": self.ids("ensemble"), "content": "Select ensemble"},
        ]

    @property
    def ensemble_selector(self) -> html.Div:
        """Dropdown to select ensemble"""
        return wcc.Dropdown(
            label="Ensemble",
            id=self.ids("ensemble"),
            options=[{"label": i, "value": i} for i in self.ensembles],
            clearable=False,
            value=self.ensembles[0],
        )

    @property
    def smry_selector(self) -> html.Div:
        """Dropdown to select ensemble"""
        return wcc.Dropdown(
            label="Time series",
            id=self.ids("vector"),
            options=[
                {
                    "label": f"{simulation_vector_description(vec)} ({vec})",
                    "value": vec,
                }
                for vec in self.smry_cols
            ],
            clearable=False,
            value=self.initial_vector,
        )

    @property
    def initial_date(self) -> datetime.date:
        df = self.smry[["ENSEMBLE", "DATE"]]
        return df.loc[df["ENSEMBLE"] == df["ENSEMBLE"].unique()[0]]["DATE"].max()

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return (
            [
                (
                    read_csv,
                    [
                        {"csv_file": self.csvfile_smry},
                        {"csv_file": self.csvfile_parameters},
                    ],
                )
            ]
            if self.csvfile_smry and self.csvfile_parameters
            else self.emodel.webvizstore
            + [
                (
                    get_realizations,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "ensemble_set_name": "EnsembleSet",
                        }
                    ],
                )
            ]
        )

    @property
    def layout(self) -> html.Div:
        return wcc.FlexBox(
            id=self.ids("layout"),
            children=[
                wcc.FlexColumn(
                    flex=1,
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        children=[
                            wcc.Selectors(
                                label="Selectors",
                                children=[self.ensemble_selector, self.smry_selector],
                            ),
                            dcc.Store(
                                id=self.ids("date-store"),
                                storage_type="session",
                            ),
                        ],
                    ),
                ),
                wcc.FlexColumn(
                    flex=3,
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        color="white",
                        highlight=False,
                        children=[
                            html.Div(
                                id=self.ids("graph-wrapper"),
                                style={"height": "450px"},
                                children=wcc.Graph(
                                    id=self.ids("graph"),
                                    clickData={"points": [{"x": self.initial_date}]},
                                ),
                            ),
                            html.Div(
                                children=[
                                    html.Div(
                                        id=self.ids("table_title"),
                                        style={"textAlign": "center"},
                                        children="",
                                    ),
                                    html.Div(
                                        style={"fontSize": "15px"},
                                        children=dash_table.DataTable(
                                            id=self.ids("table"),
                                            sort_action="native",
                                            filter_action="native",
                                            page_action="native",
                                            page_size=10,
                                        ),
                                    ),
                                ],
                            ),
                        ],
                    ),
                ),
                wcc.FlexColumn(
                    flex=3,
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        color="white",
                        highlight=False,
                        id=self.ids("tornado-wrapper"),
                        children=self.tornadoplot.layout,
                    ),
                ),
            ],
        )

    # pylint: disable=too-many-statements
    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            [
                # Output(self.ids("date-store"), "children"),
                Output(self.ids("table"), "data"),
                Output(self.ids("table"), "columns"),
                Output(self.ids("table_title"), "children"),
                Output(self.tornadoplot.storage_id, "data"),
            ],
            [
                Input(self.ids("ensemble"), "value"),
                Input(self.ids("graph"), "clickData"),
                Input(self.ids("vector"), "value"),
            ],
        )
        def _render_date(
            ensemble: str, clickdata: dict, vector: str
        ) -> Tuple[list, list, str, str]:
            """Store selected date and tornado input. Write statistics
            to table"""
            try:
                date = clickdata["points"][0]["x"]
            except TypeError as exc:
                raise PreventUpdate from exc
            data = filter_ensemble(self.smry, self.parameters, ensemble, [vector])
            data = data.loc[data["DATE"].astype(str) == date]
            table_rows, table_columns = calculate_table(data, vector)
            return (
                # json.dumps(f"{date}"),
                table_rows,
                table_columns,
                (
                    f"{simulation_vector_description(vector)} ({vector})"
                    + (
                        ""
                        if get_unit(self.smry_meta, vector) is None
                        else f" [{get_unit(self.smry_meta, vector)}]"
                    )
                ),
                json.dumps(
                    {
                        "ENSEMBLE": ensemble,
                        "data": data[["REAL", vector]].values.tolist(),
                        "number_format": "#.4g",
                        "unit": (
                            ""
                            if get_unit(self.smry_meta, vector) is None
                            else get_unit(self.smry_meta, vector)
                        ),
                    }
                ),
            )

        @app.callback(
            Output(self.ids("graph"), "figure"),
            [
                Input(self.tornadoplot.click_id, "data"),
                Input(self.tornadoplot.high_low_storage_id, "data"),
            ],
            [
                State(self.ids("ensemble"), "value"),
                State(self.ids("vector"), "value"),
                State(self.ids("graph"), "clickData"),
                State(self.ids("graph"), "figure"),
            ],
        )
        def _render_tornado(  # pylint: disable=too-many-branches, too-many-locals
            tornado_click_data_str: Union[str, None],
            high_low_storage: dict,
            ensemble: str,
            vector: str,
            date_click: dict,
            figure: dict,
        ) -> dict:
            """Update graph with line coloring, vertical line and title"""
            if callback_context.triggered is None:
                raise PreventUpdate
            ctx = callback_context.triggered[0]["prop_id"].split(".")[0]

            tornado_click: Union[dict, None] = (
                json.loads(tornado_click_data_str) if tornado_click_data_str else None
            )
            if tornado_click:
                reset_click = tornado_click["sens_name"] is None
            else:
                reset_click = False

            # Draw initial figure and redraw if ensemble/vector changes
            if ctx in ["", self.tornadoplot.high_low_storage_id] or reset_click:
                vectors = [vector]
                historical_vector_name = historical_vector(vector, self.smry_meta, True)

                if (
                    historical_vector_name is not None
                    and historical_vector_name in self.smry.columns
                ):
                    vectors.append(historical_vector_name)
                data = filter_ensemble(
                    self.smry,
                    self.parameters,
                    ensemble,
                    vectors,
                )

                line_shape = get_simulation_line_shape(
                    line_shape_fallback=self.line_shape_fallback,
                    vector=vector,
                    smry_meta=self.smry_meta,
                )
                traces = [
                    {
                        "type": "line",
                        "marker": {"color": "grey"},
                        "hoverinfo": "x+y+text",
                        "hovertext": f"Real: {r}",
                        "x": df["DATE"],
                        "y": df[vector],
                        "customdata": r,
                        "line": {"shape": line_shape},
                        "meta": {
                            "SENSCASE": df["SENSCASE"].values[0],
                            "SENSTYPE": df["SENSTYPE"].values[0],
                        },
                        "name": ensemble,
                        "legendgroup": ensemble,
                        "showlegend": r == data["REAL"][0],
                    }
                    for r, df in data.groupby(["REAL"])
                ]
                if historical_vector(vector, self.smry_meta, True) in data.columns:
                    hist = data[data["REAL"] == data["REAL"][0]]
                    traces.append(
                        {
                            "type": "line",
                            "x": hist["DATE"],
                            "y": hist[historical_vector(vector, self.smry_meta, True)],
                            "line": {
                                "shape": line_shape,
                                "color": "black",
                                "width": 3,
                            },
                            "name": "History",
                            "legendgroup": "History",
                            "showlegend": True,
                        }
                    )
                # traces[0]["hoverinfo"] = "x"
                figure = {
                    "data": traces,
                    "layout": {"margin": {"t": 60}, "hovermode": "closest"},
                }

            # Update line colors if a sensitivity is selected in tornado
            # pylint: disable=too-many-nested-blocks
            if tornado_click and tornado_click["sens_name"] in high_low_storage:
                if ctx == self.tornadoplot.high_low_storage_id:
                    tornado_click["real_low"] = high_low_storage[
                        tornado_click["sens_name"]
                    ].get("real_low")
                    tornado_click["real_high"] = high_low_storage[
                        tornado_click["sens_name"]
                    ].get("real_high")
                if reset_click:
                    add_legend = True
                    for trace in figure["data"]:
                        if trace["name"] != "History":
                            if add_legend:
                                trace["showlegend"] = True
                                add_legend = False
                            else:
                                trace["showlegend"] = False
                            trace["marker"] = {"color": "grey"}
                            trace["opacity"] = 1
                            trace["name"] = ensemble
                            trace["legendgroup"] = ensemble
                            trace["hoverinfo"] = "all"
                            trace["hovertext"] = f"Real: {trace['customdata']}"

                else:
                    add_legend_low = True
                    add_legend_high = True
                    for trace in figure["data"]:
                        if trace["name"] != "History":
                            if trace["customdata"] in tornado_click["real_low"]:
                                trace["marker"] = {
                                    "color": self.theme.plotly_theme["layout"][
                                        "colorway"
                                    ][0]
                                }
                                trace["opacity"] = 1
                                trace["legendgroup"] = "real_low"
                                trace["hoverinfo"] = "all"
                                trace["name"] = (
                                    "Below ref"
                                    if trace["meta"]["SENSTYPE"] == "mc"
                                    else trace["meta"]["SENSCASE"]
                                )
                                if add_legend_low:
                                    add_legend_low = False
                                    trace["showlegend"] = True
                                else:
                                    trace["showlegend"] = False
                            elif trace["customdata"] in tornado_click["real_high"]:
                                trace["marker"] = {
                                    "color": self.theme.plotly_theme["layout"][
                                        "colorway"
                                    ][1]
                                }
                                trace["opacity"] = 1
                                trace["legendgroup"] = "real_high"
                                trace["hoverinfo"] = "all"
                                trace["name"] = (
                                    "Above ref"
                                    if trace["meta"]["SENSTYPE"] == "mc"
                                    else trace["meta"]["SENSCASE"]
                                )
                                if add_legend_high:
                                    add_legend_high = False
                                    trace["showlegend"] = True
                                else:
                                    trace["showlegend"] = False
                            else:
                                trace["marker"] = {"color": "lightgrey"}
                                trace["opacity"] = 0.02
                                trace["showlegend"] = False
                                trace["hoverinfo"] = "skip"

            date = date_click["points"][0]["x"]
            if figure is None:
                raise PreventUpdate
            ymin = min([min(trace["y"]) for trace in figure["data"]])
            ymax = max([max(trace["y"]) for trace in figure["data"]])
            figure["layout"]["shapes"] = [
                {"type": "line", "x0": date, "x1": date, "y0": ymin, "y1": ymax}
            ]
            figure["layout"]["title"] = (
                f"Date: {date}, "
                f"Sensitivity: {tornado_click['sens_name'] if tornado_click else None}"
            )
            figure["layout"]["yaxis"] = {
                "title": f"{simulation_vector_description(vector)} ({vector})"
                + (
                    ""
                    if get_unit(self.smry_meta, vector) is None
                    else f" [{get_unit(self.smry_meta, vector)}]"
                ),
                "uirevision": vector,
            }
            figure["layout"]["xaxis"] = {"uirevision": "locked"}
            figure["layout"]["legend"] = {
                "orientation": "h",
                # "traceorder": "reversed",
                "y": 1.1,
                "x": 1,
                "xanchor": "right",
            }
            figure["layout"] = self.theme.create_themed_layout(figure["layout"])
            return figure


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def calculate_table(df: pd.DataFrame, vector: str) -> Tuple[List[dict], List[dict]]:
    table = []
    for (sensname, senscase), dframe in df.groupby(["SENSNAME", "SENSCASE"]):
        values = dframe[vector]
        try:
            table.append(
                {
                    "Sensitivity": str(sensname),
                    "Case": str(senscase),
                    "Minimum": values.min(),
                    "Maximum": values.max(),
                    "Mean": values.mean(),
                    "Stddev": values.std(),
                    "P10": np.percentile(values, 90),
                    "P90": np.percentile(values, 10),
                }
            )
        except KeyError:
            pass
    columns = [
        {**{"name": i[0], "id": i[0]}, **i[1]}
        for i in ReservoirSimulationTimeSeriesOneByOne.TABLE_STAT
    ]
    return table, columns


def filter_ensemble(
    smry: pd.DataFrame, parameters: pd.DataFrame, ensemble: str, vector: List[str]
) -> pd.DataFrame:
    smry_columns = ["DATE", "REAL"] + vector
    parameter_columns = ["REAL", "SENSCASE", "SENSNAME", "SENSTYPE"]
    return pd.merge(
        smry[smry_columns + ["ENSEMBLE"]].loc[smry["ENSEMBLE"] == ensemble][
            smry_columns
        ],
        parameters[parameter_columns + ["ENSEMBLE"]].loc[
            parameters["ENSEMBLE"] == ensemble
        ][parameter_columns],
        on=["REAL"],
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_unit(smry_meta: Union[pd.DataFrame, None], vec: str) -> Union[str, None]:
    return None if smry_meta is None else simulation_unit_reformat(smry_meta.unit[vec])



import json
from pathlib import Path
from typing import List, Type

from dash import ALL, Input, Output, callback, callback_context
from dash.development.base_component import Component
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum, callback_typecheck

from webviz_subsurface._datainput.fmu_input import find_sens_type
from webviz_subsurface._providers import EnsembleTableProviderFactory

from ._error import error
from .shared_settings import Filters, Selectors, ViewSettings
from .views.plot_view import PlotSettings, TornadoPlotView
from .views.table_view import TornadoTableView


class TornadoPlotterFMU(WebvizPluginABC):
    """Tornado plotter for FMU data from csv file og responses
    * **`ensemble`:** Which ensemble in `shared_settings` to visualize.
    * **`csvfile`:** Relative ensemble path to csv file with responses
    * **`aggregated_csvfile`:** Alternative to ensemble + csvfile with
    aggregated responses. Requires REAL and ENSEMBLE columns
    * **`aggregated_parameterfile`:** Necessary when aggregated_csvfile
    is specified. File with sensitivity specification for each realization.
    Requires columns REAL, ENSEMBLE, SENSNAME and SENSCASE.
    * **`initial_response`:** Initialize plugin with this response column
    visualized
    * **`single_value_selectors`:** List of columns in response csv file
    that should be used to select/filter data. E.g. for UNSMRY data the DATE
    column can be used. For each entry a Dropdown is shown with all unique
    values and a single value can be selected at a time.
    * **`multi_value_selectors`:** List of columns in response csv file
    to filter/select data. For each entry a Select is shown with
    all unique values. Multiple values can be selected at a time,
    and a tornado plot will be shown from the matching response rows.
    Used e.g. for volumetrics data, to select a subset of ZONES and
    REGIONS.
    """

    class Ids(StrEnum):
        TORNADO_DATA = "tornado-data"

        SELECTORS = "selectors"
        FILTERS = "filters"
        VIEW_SETTINGS = "view-settings"

        TORNADO_PLOT_VIEW = "tornado-plot-view"
        TORNADO_TABLE_VIEW = "tornado-table-view"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        webviz_settings: WebvizSettings,
        csvfile: str = None,
        ensemble: str = None,
        aggregated_csvfile: Path = None,
        aggregated_parameterfile: Path = None,
        initial_response: str = None,
        single_value_selectors: List[str] = None,
        multi_value_selectors: List[str] = None,
    ) -> None:
        super().__init__(stretch=True)

        self._single_filters = single_value_selectors if single_value_selectors else []
        self._multi_filters = multi_value_selectors if multi_value_selectors else []

        provider_factory = EnsembleTableProviderFactory.instance()
        self._error_message = ""

        self._ensemble_name = ""

        if ensemble is not None and csvfile is not None:
            ens_path = webviz_settings.shared_settings["scratch_ensembles"][ensemble]
            self._parameter_provider = (
                provider_factory.create_from_per_realization_parameter_file(ens_path)
            )
            self._table_provider = (
                provider_factory.create_from_per_realization_csv_file(ens_path, csvfile)
            )
            ensemble_name = ensemble

        elif aggregated_csvfile and aggregated_parameterfile is not None:
            self._parameter_provider = provider_factory.create_from_ensemble_csv_file(
                aggregated_parameterfile
            )
            self._table_provider = provider_factory.create_from_ensemble_csv_file(
                aggregated_csvfile
            )
            try:
                self._ensemble_name = self._table_provider.get_column_data(
                    column_names=["ENSEMBLE"]
                )
            except KeyError:
                self._error_message = f"'ENSEMBLE' is missing from {aggregated_csvfile}"

        else:
            self._error_message = (
                "Specify either ensembles and csvfile or aggregated_csvfile "
                "and aggregated_parameterfile"
            )

        try:
            design_matrix_df = self._parameter_provider.get_column_data(
                column_names=["SENSNAME", "SENSCASE"]
            )
        except KeyError:
            self._error_message = (
                "Required columns 'SENSNAME' and 'SENSCASE' is missing "
                f"from {ensemble_name}. Cannot calculate tornado plots"
            )

        design_matrix_df["ENSEMBLE"] = self._ensemble_name
        design_matrix_df["SENSTYPE"] = design_matrix_df.apply(
            lambda row: find_sens_type(row.SENSCASE), axis=1
        )

        responses: List[str] = self._table_provider.column_names()
        if self._single_filters:
            responses = [
                response
                for response in responses
                if response not in self._single_filters
            ]
        if self._multi_filters:
            responses = [
                response
                for response in responses
                if response not in self._multi_filters
            ]

        initial_response = initial_response if initial_response else responses[0]

        self.add_store(
            TornadoPlotterFMU.Ids.TORNADO_DATA,
            WebvizPluginABC.StorageType.SESSION,
        )

        self.add_shared_settings_group(
            Selectors(responses, initial_response),
            TornadoPlotterFMU.Ids.SELECTORS,
        )

        self.filters = Filters(
            self._table_provider,
            self._single_filters,
            self._multi_filters,
        )
        self.add_shared_settings_group(
            self.filters,
            TornadoPlotterFMU.Ids.FILTERS,
        )

        self.add_shared_settings_group(
            ViewSettings(design_matrix_df), TornadoPlotterFMU.Ids.VIEW_SETTINGS
        )

        self.add_view(
            TornadoPlotView(
                design_matrix_df,
                webviz_settings.theme.plotly_theme,
                TornadoPlotView.Slots(
                    reference=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.REFERENCE)
                        .to_string(),
                        "value",
                    ),
                    scale=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.SCALE)
                        .to_string(),
                        "value",
                    ),
                    filter_options=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.FILTER_OPTIONS)
                        .to_string(),
                        "value",
                    ),
                    data=Input(
                        self.get_store_unique_id(TornadoPlotterFMU.Ids.TORNADO_DATA),
                        "data",
                    ),
                    sens_filter=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.SENSITIVITIES)
                        .to_string(),
                        "value",
                    ),
                ),
            ),
            TornadoPlotterFMU.Ids.TORNADO_PLOT_VIEW,
        )

        self.add_view(
            TornadoTableView(
                design_matrix_df,
                TornadoTableView.Slots(
                    reference=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.REFERENCE)
                        .to_string(),
                        "value",
                    ),
                    scale=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.SCALE)
                        .to_string(),
                        "value",
                    ),
                    filter_options=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.FILTER_OPTIONS)
                        .to_string(),
                        "value",
                    ),
                    data=Input(
                        self.get_store_unique_id(TornadoPlotterFMU.Ids.TORNADO_DATA),
                        "data",
                    ),
                    sens_filter=Input(
                        self.shared_settings_group(TornadoPlotterFMU.Ids.VIEW_SETTINGS)
                        .component_unique_id(ViewSettings.IDs.SENSITIVITIES)
                        .to_string(),
                        "value",
                    ),
                ),
            ),
            TornadoPlotterFMU.Ids.TORNADO_TABLE_VIEW,
        )

    def _set_callbacks(self) -> None:
        @callback(
            Output(
                self.get_store_unique_id(TornadoPlotterFMU.Ids.TORNADO_DATA),
                "data",
            ),
            Input(
                self.shared_settings_group(TornadoPlotterFMU.Ids.SELECTORS)
                .component_unique_id(Selectors.IDs.RESPONSE)
                .to_string(),
                "value",
            ),
            Input(
                {
                    "id": self.filters.single_filter_id,
                    "name": ALL,
                    "type": "single_filter",
                },
                "value",
            ),
            Input(
                {
                    "id": self.filters.multi_filter_id,
                    "name": ALL,
                    "type": "multi_filter",
                },
                "value",
            ),
        )
        @callback_typecheck
        def _update_tornado_data_with_response_values(
            response: str, single_filters: list, multi_filters: list
        ) -> str:
            """Returns a json dump for the tornado data with the response values per realization"""

            data = self._table_provider.get_column_data(
                [response] + self._single_filters + self._multi_filters
            )

            # Filter data
            if single_filters is not None:
                for value, input_dict in zip(
                    single_filters, callback_context.inputs_list[1]
                ):
                    data = data.loc[data[input_dict["id"]["name"]] == value]
            if multi_filters is not None:
                for value, input_dict in zip(
                    multi_filters, callback_context.inputs_list[2]
                ):
                    data = data.loc[data[input_dict["id"]["name"]].isin(value)]

            return json.dumps(
                {
                    "ENSEMBLE": self._ensemble_name,
                    "data": data.groupby("REAL")
                    .sum()
                    .reset_index()[["REAL", response]]
                    .values.tolist(),
                    "number_format": "#.4g",
                }
            )

    @property
    def tour_steps(self) -> List[dict]:
        """Tour of the plugin"""
        tour = [
            {
                "id": self.view(TornadoPlotterFMU.Ids.TORNADO_PLOT_VIEW)
                .layout_element(TornadoPlotView.IDs.MAIN_COLUMN)
                .get_unique_id(),
                "content": ("Shows tornado plot."),
            },
            {
                "id": self.shared_settings_group(
                    TornadoPlotterFMU.Ids.SELECTORS
                ).component_unique_id(Selectors.IDs.RESPONSE),
                "content": "Choose the response for the data",
            },
        ]
        if len(self._single_filters) > 0:
            tour.append(
                {
                    "id": self.shared_settings_group(
                        TornadoPlotterFMU.Ids.FILTERS
                    ).component_unique_id(Filters.IDs.SINGLE_FILTER),
                    "content": "Choose the response for the data",
                }
            )
        if len(self._multi_filters) > 0:
            tour.append(
                {
                    "id": self.shared_settings_group(
                        TornadoPlotterFMU.Ids.FILTERS
                    ).component_unique_id(Filters.IDs.MULTI_FILTER),
                    "content": "Choose the response for the data",
                }
            )

        tour.extend(
            [
                {
                    "id": self.shared_settings_group(
                        TornadoPlotterFMU.Ids.VIEW_SETTINGS
                    ).component_unique_id(ViewSettings.IDs.REFERENCE),
                    "content": (
                        "Set reference sensitivity for which to calculate tornado plot"
                    ),
                },
                {
                    "id": self.shared_settings_group(
                        TornadoPlotterFMU.Ids.VIEW_SETTINGS
                    ).component_unique_id(ViewSettings.IDs.SCALE),
                    "content": (
                        "Set tornadoplot scale to either percentage or absolute values"
                    ),
                },
                {
                    "id": self.shared_settings_group(
                        TornadoPlotterFMU.Ids.VIEW_SETTINGS
                    ).component_unique_id(ViewSettings.IDs.SENSITIVITIES),
                    "content": ("Pick sensitivities to be displayed"),
                },
                {
                    "id": self.view(TornadoPlotterFMU.Ids.TORNADO_PLOT_VIEW)
                    .settings_group(TornadoPlotView.IDs.SETTINGS)
                    .component_unique_id(PlotSettings.IDs.PLOT_OPTIONS),
                    "content": "Options for dispaying the bars",
                },
                {
                    "id": self.view(TornadoPlotterFMU.Ids.TORNADO_PLOT_VIEW)
                    .settings_group(TornadoPlotView.IDs.SETTINGS)
                    .component_unique_id(PlotSettings.IDs.LABEL),
                    "content": "Plick settings for the label at the bars",
                },
            ]
        )
        return tour

    @property
    def layout(self) -> Type[Component]:
        return error(self._error_message)



from ._production_misfit_per_real import (
    MisfitOptions,
    MisfitPerRealView,
    PlotSettingsMisfit,
)
from ._well_production_coverage import PlotSettingsCoverage, ProdCoverageView
from ._well_production_heatmap import PlotSettingsHeatmap, ProdHeatmapView
